/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.ext.beans;

import freemarker.template.*;

import java.lang.reflect.Array;

/**
 * <p>A class that will wrap an arbitrary array into {@link TemplateCollectionModel},
 * {@link TemplateSequenceModel}, and {@link TemplateHashModel} interfaces. The models are 
 * cached (meaning requesting a model for same object twice will return the same 
 * model instance) unless the system property <tt>freemarker.beans.nocache</tt>
 * is set to true. It supports element retrieval through the <tt>array[index]</tt>
 * syntax. It also supports the <tt>array.length</tt> syntax for retrieving the length of the array, and can be
 * iterated as a list.
 * @author Attila Szegedi, <a href="mailto:szegedia@freemail.hu">szegedia@freemail.hu</a>
 * @version 1.0
 */
public final class ArrayModel
extends
    BeanModel
implements
    TemplateCollectionModel,
    TemplateSequenceModel
{
    private static final String ITERATOR = "iterator";
    private static final String LENGTH = "length";

    // Cached length of the array
    private int length;

    /**
     * Creates a new model that wraps the specified array object.
     * @param object the array object to wrap into a model.
     * @throws IllegalArgumentException if the passed object is not a Java array.
     */
    public ArrayModel(Object array)
    {
        super(array);
        Class clazz = array.getClass();
        if(!clazz.isArray())
            throw new IllegalArgumentException("Object is not an array, it is " + array.getClass().getName());
        length = Array.getLength(array);
    }

    /**
     * Returns a model wrapping the specified array object. If there is already
     * a cached model instance for this array, returns the cached model instance.
     * Models are cached using {@link WeakReference} objects. The caching can be turned
     * off by setting the <tt>freemarker.beans.nocache</tt> system property to
     * true. In this case calling this method is equivalent to constructing a new model.
     * @param object the array to wrap into a model.
     * @throws IllegalArgumentException if the passed object is not a Java array.
     * @return the model for the array
     */
    public static final BeanModel getInstance(Object object)
    {
        if(noCache)
            return new ArrayModel(object);

        BeanModelBase modelbase = lookup(object);
        if(modelbase instanceof ArrayModel)
            return (ArrayModel)modelbase;

        ArrayModel model = new ArrayModel(object);
        register(model, object);
        return model;
    }

    /**
     * Returns the type of this object (which is <tt>TYPE_ARRAY</tt>)
     */
    public int getType()
    {
        return TYPE_ARRAY;
    }

    /**
     * If the key is named "length", the length of the array is returned. Otherwise
     * acts as {@link BeanModel#get(java.lang.String)}.
     */
    public TemplateModel get(String key)
    throws
        TemplateModelException
    {
        if(LENGTH.equals(key))
            return new NumberModel(new Integer(length));
        if(ITERATOR.equals(key))
            return new Iterator();
        return super.get(key);
    }

    public TemplateModelIterator iterator()
    {
        return new Iterator();
    }

    public TemplateModel get(int index)
    throws
        TemplateModelException
    {
        try
        {
            return BeansWrapper.wrap(Array.get(object, index));
        }
        catch(IndexOutOfBoundsException e)
        {
            throw new TemplateModelException("Index out of bounds: " + index);
        }
    }

    /**
     * Returns true if the wrapped array is null, or its length is 0.
     */
    public boolean isEmpty()
    {
        return length == 0;
    }

    private class Iterator
    implements 
        TemplateSequenceModel,
        TemplateModelIterator
    {
        private int position = 0;

        public boolean isEmpty()
        {
            return ArrayModel.this.isEmpty();
        }

        public boolean hasNext()
        {
            return position < length;
        }

        public TemplateModel get(int index)
        throws
            TemplateModelException
        {
            return ArrayModel.this.get(index);
        }

        public TemplateModel next()
        throws
            TemplateModelException
        {
            return position < length ? get(position++) : null;
        }

        public int size() 
        {
            return ArrayModel.this.size();
        }
    }

    public int size() 
    {
        return length;
    }
}
