/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.ext.beans;

import freemarker.template.*;

import java.beans.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * A class that will wrap an arbitrary object into
 * {@link TemplateHashModel} interface allowing calls to arbitrary property getters
 * and invokation of accessible methods on the object from a template using the
 * <tt>object.foo</tt> to access properties and <tt>object.bar(arg1, arg2)</tt> to
 * invoke methods on it. You can also use the <tt>object.foo[index]</tt> syntax to
 * access indexed properties. It uses Beans {@link Introspector}
 * to dinamically discover the properties and methods. The models are cached (meaning requesting a model for same object
 * twice will return the same model instance) unless the system property
 * <tt>freemarker.beans.nocache</tt> is set to true.
 * @author Attila Szegedi, <a href="mailto:szegedia@freemail.hu">szegedia@freemail.hu</a>
 * @version 1.0
 */

public class BeanModel
    extends
    BeanModelBase
    implements
    TemplateHashModel
{
    private static final boolean IS_JDK14 = isJdk14();
    
    private static final Integer GENERIC_GET_KEY = new Integer(0);
    private static final String GENERIC_GET = "get";
    private static final Class[] STRING_PARAMETER = new Class[] { java.lang.String.class };
    private static final Class[] OBJECT_PARAMETER = new Class[] { java.lang.Object.class };

    // Cache of hash maps that contain already discovered properties and methods
    // for a specified class. Each key is a Class, each value is a hash map. In
    // that hash map, each key is a property/method name, each value is a
    // MethodDescriptor or a PropertyDescriptor assigned to that property/method.
    private static final Map classCache = new HashMap();
    private static Set cachedClassNames = new HashSet();

    // Cached template models that implement member properties and methods for this
    // instance. Keys are FeatureDescriptor instances (from classCache values),
    // values are either ReflectionMethodModels/ReflectionScalarModels
    private final HashMap memberMap = new HashMap();

    /**
     * Creates a new model that wraps the specified object.
     * Note that there are specialized subclasses of this class for wrapping arrays,
     * collections, enumeration, iterators, and maps. Note also that the superclass can
     * be used to wrap String objects if only scalar functionality is needed.
     * You can also choose to delegate the choice over which model class is used for
     * wrapping to {@link BeansWrapper#wrap(Object)}.
     * @param object the object to wrap into a model.
     */
    public BeanModel(Object object)
    {
        super(object);
        if (object == null) {
            return;
        }
        Class clazz = object.getClass();
        synchronized(classCache)
        {
            if(!classCache.containsKey(clazz))
            {
                String className = clazz.getName();
                if(cachedClassNames.contains(className))
                {
                    // Class reload detected, throw away caches
                    classCache.clear();
                    cachedClassNames = new HashSet();
                    BeanModelBase.clearModelCache();
                }
                classCache.put(clazz, populateClassMap(clazz));
                cachedClassNames.add(className);
            }
        }
    }

    /**
     * Returns a model wrapping the specified object. If there is already
     * a cached model instance for this object, returns the cached model instance.
     * Models are cached using {@link WeakReference} objects. The caching can be turned
     * off by setting the <tt>freemarker.beans.nocache</tt> system property to
     * true. In this case calling this method is equivalent to constructing a new model.
     * Note that there are specialized subclasses of this class for wrapping arrays,
     * collections, enumeration, iterators, and maps. Note also that the superclass can
     * be used to wrap String objects if only scalar functionality is needed.
     * You can also choose to delegate the choice over which model class is used for
     * wrapping to {@link BeansWrapper#wrap(Object)}.
     * @param object the object to wrap into a model.
     * @return the model for the object.
     */
    public static BeanModel getInstance(Object object)
    {
        if(noCache)
            return new BeanModel(object);

        BeanModel model = (BeanModel)lookup(object);
        if(model == null)
        {
            model = new BeanModel(object);
            register(model, object);
        }
        return model;
    }

    /**
     * Returns the type of this object (which is <tt>TYPE_OBJECT</tt>)
     */
    public int getType()
    {
        return TYPE_OBJECT;
    }

    /**
     * Uses Beans introspection to locate a property or method with name matching the
     * key name. If a method or property is found, it is wrapped into
     * {@link SimpleMethodModel} (for a method or indexed property), or evaluated
     * on the fly and the return value wrapped into appropriate model (for a simple property)
     * Models for various properties and methods are cached on a per-class
     * basis, so the costly introspection is performed only once per property or method
     * of a class. (Sidenote: this also implies that any class whose method has been
     * called will be strongly referred to by the framework and will not become
     * unloadable until this class has been unloaded first. Normally this is not an
     * issue, but can be in a rare scenario where you create many classes on the fly.
     * Also, as the cache grows with new classes and methods introduced to the
     * framework, it may appear as if it were leaking memory. The framework does,
     * however detect class reloads (if you happen to be in an environment that
     * does this kind of things - servlet containers do it when they reload a
     * web application) and flushes the cache.
     * If no method or propery matching the key is found, the framework
     * will try to invoke methods with signature <tt>get(java.lang.String)</tt>,
     * then <tt>get(java.lang.Object)</tt>, or alternatively (if the wrapped object
     * is a resource bundle) <tt>getObject(java.lang.String)</tt>.
     * @throws TemplateModelException if there was no property nor method nor
     * a generic <tt>get</tt> method to invoke.
     */
    public TemplateModel get(String key)
        throws
        TemplateModelException
    {
        Map keyMap = null;
        Class clazz = object.getClass();
        synchronized(classCache)
        {
            keyMap = (Map)classCache.get(clazz);
        }

        try
        {
            Object fd = keyMap.get(key);

            return
                fd == null
                ? invokeGenericGet(keyMap, clazz, key)
                : invokeThroughDescriptor(fd);
        }
        catch(TemplateModelException e)
        {
            throw e;
        }
        catch(Exception e)
        {
            java.io.StringWriter sw = new java.io.StringWriter();
            java.io.PrintWriter pw = new java.io.PrintWriter(sw);
            pw.println("Failed on " + object.getClass().getName());
            e.printStackTrace(pw);
            throw new TemplateModelException(sw.toString());
        }
    }

    /**
     * Populates a map with property and method descriptors for a specified
     * class. If any property or method descriptors specifies a read method
     * that is not accessible, replaces it with appropriate accessible method
     * from a superclass or interface. Introspection exceptions are dumped to
     * System.err and otherwise ignored, as there is no standard logging
     * facility in Freemarker.
     */
    private static Map populateClassMap(Class clazz)
    {
        // Populate first from bean info
        Map map = populateClassMapWithBeanInfo(clazz);
        // Next add the generic "get" method
        Method genericGet = findGenericGetMethod(clazz);
        if(genericGet != null)
        {
            if(map.isEmpty())
                map = Collections.singletonMap(GENERIC_GET_KEY, genericGet);
            else
                map.put(GENERIC_GET_KEY, genericGet);
        }
        return map;
    }

    private static Map populateClassMapWithBeanInfo(Class clazz)
    {
        try
        {
            BeanInfo beanInfo = Introspector.getBeanInfo(clazz);
            PropertyDescriptor[] pda = beanInfo.getPropertyDescriptors();
            MethodDescriptor[] mda = beanInfo.getMethodDescriptors();

            int memberCount = pda.length + mda.length;
            Map classMap = new HashMap(memberCount * 4/3, .75f);
            boolean dontCalculateAccessibles = IS_JDK14 || Modifier.isPublic(clazz.getModifiers());
			Map accessibleMethods = dontCalculateAccessibles ? null : getAccessibleMethods(clazz);
			
            for(int i = pda.length - 1; i >= 0; --i)
            {
                PropertyDescriptor pd = pda[i];
                if(pd instanceof IndexedPropertyDescriptor)
                {
                    IndexedPropertyDescriptor ipd = (IndexedPropertyDescriptor)pd;
                    Method readMethod = ipd.getIndexedReadMethod();
                    Method publicReadMethod = getAccessibleMethod(readMethod, accessibleMethods);
                    if(publicReadMethod != null)
                    {
                        try
                        {
                            if(readMethod != publicReadMethod)
                                ipd.setIndexedReadMethod(publicReadMethod);
                            classMap.put(ipd.getName(), ipd);
                        }
                        catch(IntrospectionException e)
                        {
                            e.printStackTrace();
                        }
                    }
                }
                else
                {
                    Method readMethod = pd.getReadMethod();
                    Method publicReadMethod = getAccessibleMethod(readMethod, accessibleMethods);
                    if(publicReadMethod != null)
                    {
                        try
                        {
                            if(readMethod != publicReadMethod)
                                pd.setReadMethod(publicReadMethod);
                            classMap.put(pd.getName(), pd);
                        }
                        catch(IntrospectionException e)
                        {
                            e.printStackTrace();
                        }
                    }
                }
            }
            for(int i = mda.length - 1; i >= 0; --i)
            {
                MethodDescriptor md = mda[i];
                Method method = md.getMethod();
                Method publicMethod = getAccessibleMethod(method, accessibleMethods);
                if(publicMethod != null)
                {
                    String name = md.getName();
                    Object previous = classMap.get(name);
                    if(previous instanceof Method)
                    {
                        // Overloaded method - replace method with a method map
                        MethodMap methodMap = new MethodMap(name);
                        methodMap.addMethod((Method)previous);
                        methodMap.addMethod(publicMethod);
                        classMap.put(name, methodMap);
                    }
                    else if(previous instanceof MethodMap)
                    {
                        // Already overloaded method - add new overload
                        ((MethodMap)previous).addMethod(publicMethod);
                    }
                    else
                    {
                        // Simple method (this far)
                    	classMap.put(name, publicMethod);
                    }
                }
            }
            return classMap;
        }
        catch(IntrospectionException e)
        {
            e.printStackTrace();
            return Collections.EMPTY_MAP;
        }
    }

	private static Method getAccessibleMethod(Method m, Map accessibles)
	{
	    if(accessibles == null || Modifier.isPublic(m.getDeclaringClass().getModifiers()))
	    {
	        return m;
	    }
	    return (Method)accessibles.get(m);
	}
	
    private static Method findGenericGetMethod(Class clazz)
    {
        Method genericGet = null;
        try
        {
            // First try "get(java.lang.String)"
            genericGet = BeansWrapper.getPublicMethod(clazz.getDeclaredMethod(GENERIC_GET, STRING_PARAMETER));
        }
        catch(NoSuchMethodException e)
        {
            // Not found...
        }

        if(genericGet == null)
        {
            try
            {
                // Next try "get(java.lang.Object)"
                genericGet = BeansWrapper.getPublicMethodOrNull(clazz.getDeclaredMethod(GENERIC_GET, OBJECT_PARAMETER));
            }
            catch(NoSuchMethodException e)
            {
                // Not found...
            }
        }
        return genericGet;
    }

    private TemplateModel invokeThroughDescriptor(Object desc)
        throws
        NoSuchMethodException,
        IllegalAccessException,
        InvocationTargetException
    {
        // See if this particular instance has a cached implementation
        // for the requested feature descriptor
        TemplateModel member = null;
        synchronized(memberMap)
        {
            member = (TemplateModel)memberMap.get(desc);
        }

        if(member != null)
            return member;

        TemplateModel retval = null;
        if(desc instanceof IndexedPropertyDescriptor)
        {
            retval = member = new SimpleMethodModel(object, ((IndexedPropertyDescriptor)desc).getIndexedReadMethod());
        }
        else if(desc instanceof PropertyDescriptor)
        {
            PropertyDescriptor pd = (PropertyDescriptor)desc;
            retval = BeansWrapper.wrap(pd.getReadMethod().invoke(object, null));
            // (member == null) condition remains, as we don't cache these
        }
        else if(desc instanceof Method)
        {
            retval = member = new SimpleMethodModel(object, (Method)desc);
        }
        else if(desc instanceof MethodMap)
        {
            retval = member = new OverloadedMethodModel(object, (MethodMap)desc);
        }
        
        // If new cacheable member was created, cache it
        if(member != null)
        {
            synchronized(memberMap)
            {
                memberMap.put(desc, member);
            }
        }
        return retval;
    }

    protected TemplateModel invokeGenericGet(Map keyMap, Class clazz, String key)
    throws
        IllegalAccessException,
        InvocationTargetException,
        TemplateModelException
    {
        Method genericGet = (Method)keyMap.get(GENERIC_GET_KEY);
        if(genericGet == null)
            throw new TemplateModelException("No such key: " + key + " on instance of " + object.getClass().getName() + " " + keyMap);

        return BeansWrapper.wrap(genericGet.invoke(object, new Object[] { key }));
    }
    
    
    /**
     * Retrieves mapping of methods to accessible methods for a class.
     * In case the class is not public, retrieves methods with same 
     * signature as its public methods from public superclasses and 
     * interfaces (if they exist). Basically upcasts every method to the 
     * nearest acccessible method.
     */
    private static Map getAccessibleMethods(Class clazz)
    {
        Method[] methods = clazz.getMethods();
        
        /*
         *  Short circuit for the (hopefully) majority of cases where the
         *  clazz is public
         */
        
        if (Modifier.isPublic(clazz.getModifiers()))
        {
            return createIdentityMap(methods);
        }

        /*
         *  No luck - the class is not public, so we're going the longer way.
         */

        MethodInfo[] methodInfos = new MethodInfo[methods.length];

        for(int i = methods.length; i-- > 0; )
        {
            methodInfos[i] = new MethodInfo(methods[i]);
        }

        int upcastCount = getAccessibleMethods(clazz, methodInfos, 0);

        Map map = new HashMap(upcastCount * 4/3, .75f);
        for(int i = 0; i < methodInfos.length; ++i)
        {
            MethodInfo methodInfo = methodInfos[i];
            if(methodInfo.upcast)
            {
                map.put(methodInfo.oldMethod, methodInfo.method);
            }
        }
        return map;
    }

    /**
     *  Recursively finds a match for each method, starting with the class, and then
     *  searching the superclass and interfaces.
     *
     *  @param clazz Class to check
     *  @param methodInfos array of methods we are searching to match
     *  @param upcastCount current number of methods we have matched
     *  @return count of matched methods
     */
    private static int getAccessibleMethods(Class clazz, MethodInfo[] methodInfos, int upcastCount)
    {
        int l = methodInfos.length;
        
        /*
         *  if this class is public, then check each of the currently
         *  'non-upcasted' methods to see if we have a match
         */

        if( Modifier.isPublic(clazz.getModifiers()) )
        {
            for(int i = 0; i < l && upcastCount < l; ++i)
            {
                try
                {
                    MethodInfo methodInfo = methodInfos[i];

                    if(!methodInfo.upcast)
                    {
                        methodInfo.tryUpcasting(clazz);
                        upcastCount++;
                    }
                }
                catch(NoSuchMethodException e)
                {
                    /*
                     *  Intentionally ignored - it means
                     *  it wasn't found in the current class
                     */
                }
            }

            /*
             *  Short circuit if all methods were upcast
             */

            if(upcastCount == l)
            {
                return upcastCount;
            }
        }

        /*
         *   Examine superclass
         */

        Class superclazz = clazz.getSuperclass();

        if(superclazz != null)
        {
            upcastCount = getAccessibleMethods(superclazz , methodInfos, upcastCount);

            /*
             *  Short circuit if all methods were upcast
             */

            if(upcastCount == l)
            {
                return upcastCount;
            }
        }

        /*
         *  Examine interfaces. Note we do it even if superclazz == null.
         *  This is redundant as currently java.lang.Object does not implement
         *  any interfaces, however nothing guarantees it will not in future.
         */

        Class[] interfaces = clazz.getInterfaces();

        for(int i = interfaces.length; i-- > 0; )
        {
            upcastCount = getAccessibleMethods(interfaces[i], methodInfos, upcastCount);

            /*
             *  Short circuit if all methods were upcast
             */

            if(upcastCount == l)
            {
                return upcastCount;
            }
        }

        return upcastCount;
    }

    private static Map createIdentityMap(Object[] elements)
    {
        Map map = new HashMap(elements.length * 4/3, .75f);
        for(int i = elements.length; i-->0;)
            map.put(elements[i], elements[i]);
        return map;
    }

    /**
     *  Used for the iterative discovery process for public methods.
     */
    private static final class MethodInfo
    {
        Method oldMethod;
        Method method;
        String name;
        Class[] parameterTypes;
        boolean upcast;
        
        MethodInfo(Method method)
        {
            oldMethod = method;
            this.method = null;
            name = method.getName();
            parameterTypes = method.getParameterTypes();
            upcast = false;
        }
        
        void tryUpcasting(Class clazz)
            throws NoSuchMethodException
        {
            method = clazz.getMethod(name, parameterTypes);
            name = null;
            parameterTypes = null;
            upcast = true;
        }
    }
    
    private static final boolean isJdk14()
    {
        try
        {
            Class.forName("java.nio.ByteBuffer");
            return true;
        }
        catch(Exception e)
        {
            return false;
        }
        
    }
}
