/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.ext.beans;

import freemarker.ext.util.IdentityHashMap;

import freemarker.template.*;
import java.lang.ref.*;
import java.util.*;

/**
 * Base class for all classes that wrap objects into template models.
 * Instances of this class and descendant classes can be created directly,
 * or utilizing a model cache through <tt>getInstance()</tt> calls, or polymorphically
 * through the {@link BeansWrapper#wrap(Object)} method. This class basically
 * just provides mapping from <tt>Object.toString()</tt> to <tt>TemplateScalarModel#getAsString(java.util.Locale)</tt>
 * @author Attila Szegedi, <a href="mailto:szegedia@freemail.hu">szegedia@freemail.hu</a>
 * @version 1.0
 */
class BeanModelBase
implements
	TemplateScalarModel
{
	/** Returned from {@link #getType()} if the object is a BeanModelBase */
	public static final int TYPE_OBJECT_BASE = -1;
	/** Returned from {@link #getType()} if the object is a BeanModel */
	public static final int TYPE_OBJECT = 0;
	/** Returned from {@link #getType()} if the object is a ArrayModel */
	public static final int TYPE_ARRAY = 1;
	/** Returned from {@link #getType()} if the object is a CollectionModel */
	public static final int TYPE_COLLECTION = 2;
	/** Returned from {@link #getType()} if the object is a MapModel */
	public static final int TYPE_MAP = 3;
	/** Returned from {@link #getType()} if the object is a IteratorModel */
	public static final int TYPE_ITERATOR = 4;
	/** Returned from {@link #getType()} if the object is a EnumerationModel */
	public static final int TYPE_ENUMERATION = 5;
	/** Returned from {@link #getType()} if the object is a ResourceBundleModel */
	public static final int TYPE_RESOURCE_BUNDLE = 6;

	static final boolean noCache = Boolean.getBoolean("freemarker.beans.nocache") || Boolean.getBoolean("expose.reflection.nocache");
        // This is a DEBUG flag, that should be set to true only in builds that
        // are used to tune performance
        private static final boolean trackCaching = false;
        static int cacheHits = 0;
        static int cacheMisses = 0;

        private static final Map modelCache = noCache ? null : new IdentityHashMap();
	private static final ReferenceQueue refQueue = noCache ? null : new ReferenceQueue();

	Object object;

	BeanModelBase(Object object)
	{
		this.object = object;
	}

	/**
	 * Creates a new model that wraps the specified String object.
	 * Note that since the only Freemarker-specific method in this class is
	 * <tt>getAsString</tt> there is no sense in creating an object of this
	 * class directly with anything except a String. If you need to wrap an object
	 * of another class, use one of specialized subclasses, or delegate the choice
	 * to {@link BeansWrapper#wrap(Object)}.
	 * @param object the String to wrap into a model.
	 */
	BeanModelBase(String object)
	{
		this.object = object;
	}

	/**
	 * Returns a model wrapping the specified String object. If there is already
	 * a cached model instance for this String, returns the cached model instance.
	 * Models are cached using {@link WeakReference} objects. The caching can be turned
	 * off by setting the <tt>freemarker.beans.nocache</tt> system property to
	 * true. In this case calling this method is equivalent to constructing a new model.
	 * Note that since the only Freemarker-specific method in this class is
	 * <tt>getAsString</tt> there is no sense in creating an object of this
	 * class directly with anything except a String. If you need to wrap an object
	 * of another class, use one of specialized subclasses, or delegate the choice
	 * to {@link BeansWrapper#wrap(Object)}.
	 * @param object the String to wrap into a model.
	 * @return the model for the String
	 */
	public static final BeanModelBase getInstance(String object)
	{
		if(noCache)
                {
                    if(trackCaching)
                        doTrackCaching(false);
		    return new BeanModelBase(object);
                }

		BeanModelBase model = lookup(object);
		if(model == null)
		{
                    if(trackCaching)
                        doTrackCaching(false);
		    model = new BeanModelBase(object);
		    register(model, object);
		}
                else if(trackCaching)
                    doTrackCaching(true);

		return model;
	}

        private static synchronized void doTrackCaching(boolean cacheHit)
        {
            if(cacheHit)
                cacheHits++;
            else
                cacheMisses++;
            if((cacheHits + cacheMisses) % 100 == 0)
                System.err.println("Expose cache: hits=" + cacheHits + " misses=" + cacheMisses);
        }

        /**
         * Forces clearing of model cache.
         */
        static void clearModelCache()
        {
            synchronized(modelCache)
            {
                modelCache.clear();
            }
        }

	/**
	 * Returns the wrapped object.
	 */
	public Object getObject()
	{
		return object;
	}

	final static BeanModelBase lookup(Object object)
	{
		ModelReference ref = null;
		// NOTE: we're doing minimal synchronizations -- which can lead to
		// duplicate wrapper creation. However, this has no harmful side-effects and
		// is a lesser performance hit.
		synchronized(modelCache)
		{
			ref = (ModelReference)modelCache.get(object);
		}

		if(ref != null)
			return ref.getModel();

		return null;
	}

	final static void register(BeanModelBase model, Object object)
	{
		synchronized(modelCache)
		{
			// Remove cleared references
			for(;;)
			{
				ModelReference queuedRef = (ModelReference)refQueue.poll();
				if(queuedRef == null)
					break;
				modelCache.remove(queuedRef.object);
			}
			// Register new reference
			modelCache.put(object, new ModelReference(model));
		}
	}

	/**
	 * Returns the type of this object (which is <tt>TYPE_OBJECT_BASE</tt>)
	 */
	public int getType()
	{
		return TYPE_OBJECT_BASE;
	}

	/**
	 * Returns either the toString() of the wrapped object or the string "null",
	 * if the wrapped object is null.
         * @param l unused
	 */
	public String getAsString(Locale l)
	{
		return object == null ? "null" : object.toString();
	}

	/**
	 * Tells whether the model is empty. It is empty if either the wrapped object is
	 * null, or it is a Boolean with false value.
	 */
	public boolean isEmpty()
	{
            return object == null || Boolean.FALSE.equals(object);
	}

	/**
	 * A special soft reference that is registered in the modelCache.
	 * When it gets cleared (that is, the model became unreachable)
	 * it will remove itself from the model cache.
	 */
	private static final class ModelReference extends SoftReference
	{
		Object object;

		ModelReference(BeanModelBase ref)
		{
			super(ref, refQueue);
			object = ref.object;
		}

		BeanModelBase getModel()
		{
			BeanModelBase model = (BeanModelBase)this.get();
			return model;
		}
	}

	private static String print(Object object)
	{
		if(object == null)
			return null;
		return object.getClass().getName() + "@" + Integer.toHexString(System.identityHashCode(object));
	}
}
