/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.ext.beans;

import freemarker.template.*;
import java.beans.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * Utility class that provides generic services to reflection classes.
 * It handles all polymorphism issues in the {@link #wrap(Object)} and {@link #unwrap(Object)} methods.
 * @author Attila Szegedi, <a href="mailto:szegedia@freemail.hu">szegedia@freemail.hu</a>
 * @version 1.0
 */
public final class BeansWrapper
{
    private static final Map publicMethodCache = new WeakHashMap();

    private BeansWrapper() {}

    /**
     * Wraps the object with a template model that is most specific for the object's
     * class. Specifically:
     * <ul>
     * <li>if the object is null, returns {@link Models#EMPTY_SCALAR},</li>
     * <li>if the object is a String returns a {@link BeanModelBase} for it,</li>
     * <li>if the object is a Number returns a {@link NumberModel} for it,</li>
     * <li>if the object is a Boolean returns a {@link BooleanScalar} for it,</li>
     * <li>if the object is already a TemplateModel, returns it unchanged,</li>
     * <li>if the object is an array, returns a {@link ArrayModel} for it
     * <li>if the object is a Map, returns a {@link MapModel} for it
     * <li>if the object is a Collection, returns a {@link CollectionModel} for it
     * <li>if the object is an Iterator, returns a {@link IteratorModel} for it
     * <li>if the object is an Enumeration, returns a {@link EnumerationModel} for it
     * <li>otherwise, returns a generic {@link BeanModel} for it.
     * </ul>
     */
    public static final TemplateModel wrap(Object object)
    {
        if(object == null)
            return Models.EMPTY_SCALAR;
        if(object instanceof TemplateModel)
            return (TemplateModel)object;
        if(object.getClass().isArray())
            return ArrayModel.getInstance(object);
        if(object instanceof Map)
            return MapModel.getInstance((Map)object);
        if(object instanceof Collection)
            return CollectionModel.getInstance((Collection)object);
        if(object instanceof Number)
            return NumberModel.getInstance((Number)object);
        if(object instanceof ResourceBundle)
            return ResourceBundleModel.getInstance((ResourceBundle)object);
        if(object instanceof Boolean)
            return BooleanScalar.getInstance((Boolean)object);
        if(object instanceof Iterator)
            return IteratorModel.getInstance((Iterator)object);
        if(object instanceof Enumeration)
            return EnumerationModel.getInstance((Enumeration)object);
        return BeanModel.getInstance(object);
    }

    /**
     * Attempts to unwrap a model into underlying object. It can unwrap
     * {@link BeanModelBase} instances, as well as generic
     * {@link TemplateNumberModel} and {@link TemplateScalarModel} instances.
     * All other objects are returned unchanged.
     */
    public static final Object unwrap(Object model)
        throws
        TemplateModelException
    {
        if(model instanceof BeanModelBase)
            return ((BeanModelBase)model).getObject();
        if(model instanceof TemplateNumberModel)
            return ((TemplateNumberModel)model).getAsNumber();
        if(model instanceof TemplateScalarModel)
            return ((TemplateScalarModel)model).getAsString(Locale.getDefault());
        return model;
    }

    /**
     * For a given method, retrieves its publicly accessible counterpart. If a
     * public method is declared by a non-public class, calling it throws an
     * IllegalAccessException. This method will look for a method with same name
     * and signature declared in a public superclass or implemented interface of this
     * method's declaring class. This counterpart method is publicly callable.
     * @param method a method whose publicly callable counterpart is requested.
     * @return the publicly callable counterpart method. Note that if the parameter
     * method is itself declared by a public class, this method is an identity
     * function.
     * @throws NoSuchMethodException if an appropriate method could not be located.
     */
    static final Method getPublicMethod(Method method)
        throws
        NoSuchMethodException
    {
        Class clazz = method.getDeclaringClass();
        // Short circuit for (hopefully the majority of) cases where the declaring
        // class is public.
        if((clazz.getModifiers() & Modifier.PUBLIC) != 0)
            return method;

        synchronized(publicMethodCache)
        {
            // Look first in the cache
            Method pubmethod = (Method)publicMethodCache.get(method);
            // If not cached ...
            if(pubmethod == null)
            {
                // ...lookup, and ...
                pubmethod = getPublicMethod(clazz, method.getName(), method.getParameterTypes());
                // ...cache.
                publicMethodCache.put(method, pubmethod);
            }
            return pubmethod;
        }
    }

    static final Method getPublicMethodOrNull(Method method)
    {
        if(method == null)
            return null;
        try
        {
            return getPublicMethod(method);
        }
        catch(NoSuchMethodException e)
        {
            return null;
        }
    }

    /**
     * Looks up the method with specified name and signature in the first public
     * superclass or implemented interface of the class. This method is safe to call
     * from reflection, and will not throw IllegalAccessException.
     * @param class the class whose method is sought
     * @param name the name of the method
     * @param paramTypes the classes of method parameters
     */
    private final static Method getPublicMethod(Class clazz, String name, Class[] paramTypes)
        throws
        NoSuchMethodException
    {
        if((clazz.getModifiers() & Modifier.PUBLIC) != 0)
            return clazz.getMethod(name, paramTypes);

        Class superclazz = clazz.getSuperclass();
        if(superclazz != null)
        {
            try
            {
                return getPublicMethod(superclazz, name, paramTypes);
            }
            catch(NoSuchMethodException e)
            {
                // Intentionally ignored
            }
        }

        Class[] interfaces = clazz.getInterfaces();
        for(int i = 0; i < interfaces.length; ++i)
        {
            try
            {
                return getPublicMethod(interfaces[i], name, paramTypes);
            }
            catch(NoSuchMethodException e)
            {
                // Intentionally ignored
            }
        }

        throw new NoSuchMethodException();
    }
}
