/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.ext.beans;

import freemarker.template.*;

/**
 * Provides singleton template models for representing boolean values.
 * @author Attila Szegedi, <a href="mailto:szegedia@freemail.hu">szegedia@freemail.hu</a>
 * @version 1.0
 */

public final class BooleanScalar
extends
    BeanModel
{
	/**
	 * TemplateScalarModel for representing boolean true value
	 */
	public static final BooleanScalar TRUE = new BooleanScalar(true);
	/**
	 * TemplateScalarModel for representing boolean false value
	 */
	public static final BooleanScalar FALSE = new BooleanScalar(false);

	private Boolean value;

	private BooleanScalar(boolean value)
	{
		super(value ? Boolean.TRUE : Boolean.FALSE);
	}

	/**
	 * Returns the boolean singleton TemplateScalarModel appropriate for
	 * representing the specified boolean value.
	 */
	public static final BooleanScalar getInstance(boolean value)
	{
		return value ? TRUE : FALSE;
	}

	/**
	 * Returns the boolean singleton TemplateScalarModel appropriate for
	 * representing the specified boolean value.
	 */
	public static final BooleanScalar getInstance(Boolean value)
	{
		return value.booleanValue() ? TRUE : FALSE;
	}

	/**
	 * Returns true if the passed object is the TRUE singleton.
	 */
	public static final boolean getBoolean(TemplateModel value)
	{
		return value == TRUE;
	}

	/**
	 * Returns {@link Boolean#TRUE} if the passed object is the TRUE singleton,
	 * otherwise returns {@link Boolean#FALSE}.
	 */
	public static final Boolean getBooleanObject(Object value)
	{
		return value == TRUE ? Boolean.TRUE : Boolean.FALSE;
	}

	/**
	 * Returns true if this is the FALSE object. As the <tt>Expression.isTrue()</tt> returns true if
	 * the model is </tt>!isEmpty()</tt>, this allows for usual semantics of
	 * booleans (you can write <tt>&lt;if cond></tt> instead of
	 * <tt>&lt;if cond == "true"></tt>).
	 */
	public boolean isEmpty()
	{
		return this == FALSE;
	}
}
