/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.ext.beans;

import freemarker.template.*;
import java.util.*;

/**
 * <p>A special case of {@link BeanModel} that can wrap Java collections
 * and that implements the {@link TemplateCollectionModel} in order to be iterable 
 * in a <tt>&lt;list></tt> block.</p>
 * @author Attila Szegedi, <a href="mailto:szegedia@freemail.hu">szegedia@freemail.hu</a>
 * @version 1.0
 */
public final class CollectionModel
extends
    BeanModel
implements
    TemplateCollectionModel,
    TemplateSequenceModel
{
    private static final String ITERATOR = "iterator";

    /**
     * Creates a new model that wraps the specified collection object.
     * @param object the collection object to wrap into a model.
     */
    public CollectionModel(Collection collection)
    {
        super(collection);
    }

    /**
     * Returns a model wrapping the specified collection object. If there is already
     * a cached model instance for this collection, returns the cached model instance.
     * Models are cached using {@link WeakReference} objects. The caching can be turned
     * off by setting the <tt>freemarker.beans.nocache</tt> system property to
     * true. In this case calling this method is equivalent to constructing a new model.
     * @param object the collection to wrap into a model.
     * @return the model for the collection
     */
    public static final CollectionModel getInstance(Collection object)
    {
        if(noCache)
            return new CollectionModel(object);

        CollectionModel model = (CollectionModel)lookup(object);
        if(model == null)
        {
            model = new CollectionModel(object);
            register(model, object);
        }
        return model;
    }

    public TemplateModel get(String key)
        throws
        TemplateModelException
    {
        return ITERATOR.equals(key) ? new IteratorModel(((Collection)object).iterator()) : super.get(key);
    }

    /**
     * Retrieves the i-th object from the collection, wrapped as a TemplateModel.
     * @throws TemplateModelException if the index is out of bounds, or the
     * underlying collection is not a List.
     */
    public TemplateModel get(int index)
    throws
        TemplateModelException
    {
        if(object instanceof List)
        {
            try
            {
                return BeansWrapper.wrap(((List)object).get(index));
            }
            catch(IndexOutOfBoundsException e)
            {
                throw new TemplateModelException("Index out of bounds: " + index);
            }
        }
        else
        {
            throw new TemplateModelException("Underlying collection is not a list, it's " + object.getClass().getName());
        }
    }
    
    public TemplateModelIterator iterator()
    {
        return new IteratorModel(((Collection)object).iterator());
    }
    
    /**
     * Returns the type of this object (which is <tt>TYPE_COLLECTION</tt>)
     */
    public int getType()
    {
        return TYPE_COLLECTION;
    }

    /**
     * Returns true if the underlying collection contains no elements.
     */
    public boolean isEmpty()
    {
        return ((Collection)object).isEmpty();
    }

    public int size()
    {
        return ((Collection)object).size();
    }
}
