/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.ext.beans;

import freemarker.template.*;
import java.util.*;

/**
 * <p>A class that adds {@link TemplateModelIterator} functionality to the
 * {@link Iterator} interface implementers. Also, it implements {@link TemplateListModel}
 * because of certain backward-compatibility reasons.</p>
 * <p>It differs from the {@link SimpleIterator} in that it inherits from
 * {@link BeanModel}, and therefore you can call methods on it directly,
 * even to the effect of calling <tt>iterator.remove()</tt> in the template.</p>
 * <p>Using the model as a list model is NOT thread-safe, as iterators are
 * inherently not thread-safe.</p>
 * @author Attila Szegedi, <a href="mailto:szegedia@freemail.hu">szegedia@freemail.hu</a>
 * @version 1.0
 */

public final class IteratorModel
extends
    BeanModel
implements
    TemplateModelIterator
{
    /**
     * Creates a new model that wraps the specified iterator object.
     * @param object the iterator object to wrap into a model.
     */
    public IteratorModel(Iterator iterator)
    {
        super(iterator);
    }

    /**
     * Returns a model wrapping the specified iterator object. It always returns a new
     * object, as Iterators are considered short-lived and caching them would have little
     * sense.
     * @return the model for the iterator
     */
    public static IteratorModel getInstance(Iterator object)
    {
        return new IteratorModel(object);
    }

    /**
     * Returns the type of this object (which is <tt>TYPE_ITERATOR</tt>)
     */
    public int getType()
    {
        return TYPE_ITERATOR;
    }

    /**
     * Calls {@link Iterator#hasNext()}.
     */
    public boolean hasNext()
    {
        return ((Iterator)object).hasNext();
    }

    /**
     * Calls {@link Iterator#next()}.
     */
    public TemplateModel next()
    {
        Iterator i = (Iterator)object;
        return i.hasNext() ? BeansWrapper.wrap(i.next()) : null;
    }
}
