/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.ext.beans;

import freemarker.template.*;
import java.util.*;

/**
 * <p>A special case of {@link BeanModel} that adds implementation
 * for {@link TemplateMethodModelEx} on map objects that is a shortcut for the
 * <tt>Map.get()</tt> method. Note that if the passed argument itself is a
 * reflection-wrapper model, then the map lookup will be performed using the
 * wrapped object as the key. Note that you can call <tt>get()</tt> using the
 * <tt>map.key</tt> syntax inherited from {@link BeanModel} as well, 
 * however in that case the key is always a string.</p>
 * <p>The class itself does not implement the {@link TemplateCollectionModel}.
 * You can, however use <tt>map.entrySet()</tt>, <tt>map.keySet()</tt>, or
 * <tt>map.values()</tt> to obtain {@link TemplateCollectionModel} instances for 
 * various aspects of the map.</p>
 * @author Attila Szegedi, <a href="mailto:szegedia@freemail.hu">szegedia@freemail.hu</a>
 * @version 1.0
 */
public class MapModel
    extends
    BeanModel
    implements
    TemplateMethodModelEx
{
    private ThreadLocal iterator = new ThreadLocal();

    /**
     * Creates a new model that wraps the specified map object.
     * @param object the map object to wrap into a model.
     */
    public MapModel(Map map)
    {
        super(map);
    }

    /**
     * Returns a model wrapping the specified map object. If there is already
     * a cached model instance for this map, returns the cached model instance.
     * Models are cached using {@link WeakReference} objects. The caching can be turned
     * off by setting the <tt>freemarker.beans.nocache</tt> system property to
     * true. In this case calling this method is equivalent to constructing a new model.
     * @param object the map to wrap into a model.
     * @return the model for the map
     */
    public static final MapModel getInstance(Map object)
    {
        if(noCache)
            return new MapModel(object);

        MapModel model = (MapModel)lookup(object);
        if(model == null)
        {
            model = new MapModel(object);
            register(model, object);
        }
        return model;
    }

    /**
     * Returns the type of this object (which is <tt>TYPE_MAP</tt>)
     */
    public int getType()
    {
        return TYPE_MAP;
    }

    /**
     * The first argument is used as a key to call the map's <tt>get</tt> method.
     */
    public TemplateModel exec(List arguments)
    throws
        TemplateModelException
    {
        Object key = BeansWrapper.unwrap(arguments.get(0));
        return BeansWrapper.wrap(((Map)object).get(key));
    }

    /**
     * Overridden to invoke the generic get method by casting to Map instead of 
     * through reflection - should yield better performance.
     */
    protected TemplateModel invokeGenericGet(Map keyMap, Class clazz, String key)
    {
        return BeansWrapper.wrap(((Map)object).get(key));
    }

    /**
     * Returns true if the underlying map is empty.
     */
    public boolean isEmpty()
    {
        return ((Map)object).isEmpty();
    }
}
