/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.ext.beans;

import freemarker.template.*;
import java.util.*;

/**
 * A wrapper for an arbitrary map that will expose it as a {@link TemplateModelRoot}.
 * All values returned from the {@link #get(java.lang.String)} method are automatically
 * passed through {@link BeansWrapper#wrap(java.lang.Object)}.
 * @author Attila Szegedi, Scriptum Inc.
 * @version 1.0
 */
public class RootMapModel
implements
	TemplateModelRoot
{
	private Map map;
        private Locale locale;

	/**
	 * Creates a new model with its own internal map.
	 */
	public RootMapModel()
	{
		map = new HashMap();
	}

	/**
	 * Creates a new model that adds {@link TemplateModelRoot} behavior to the
	 * passed map.
	 */
	public RootMapModel(Map map)
	{
		this.map = map;
	}

	/**
	 * Returns <tt>map.isEmpty()</tt> where <tt>map</tt> is the underlying map.
	 */
	public boolean isEmpty()
	{
		return map.isEmpty();
	}

	/**
	 * Returns <tt>BeansWrapper.wrap(map.get(key))</tt> where <tt>map</tt> is the underlying map.
	 */
	public TemplateModel get(String key)
	{
		return BeansWrapper.wrap(map.get(key));
	}

	/**
	 * Puts the specified template model into the map.
	 */
	public void put(String key, TemplateModel model)
	{
		map.put(key, model);
	}

	/**
	 * Removes the template model with the specified key from the map.
	 */
	public void remove(String key)
	{
		map.remove(key);
	}

// added these methods so that it can still implement TemplateModelRoot
// since I fattened the interface (revusky)

        public void setLocale(Locale locale) 
        {
            this.locale = locale;
        }

        public Locale getLocale() 
        {
            if (locale == null) 
            {
                locale = Locale.getDefault();
            }
            return locale;
        }
}
