/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.ext.beans;

import freemarker.template.*;

import java.lang.reflect.*;
import java.math.BigDecimal;
import java.io.*;
import java.util.*;

/**
 * A class that will wrap a reflected method call into a
 * {@link freemarker.template.TemplateMethodModel} interface. 
 * It is used by {@link BeanModel} to wrap reflected method calls
 * for non-overloaded methods.
 * @author Attila Szegedi, szegedia at users dot sourceforge dot net
 * @version 1.0
 */
final class SimpleMethodModel
    implements
    TemplateMethodModelEx,
    TemplateSequenceModel
{
    private final Object object;
    private final Method method;

    /**
     * Creates a model for a specific method on a specific object.
     * @param object the object to call the method on. Can be
     * <tt>null</tt> for static methods.
     * @param method the method that will be invoked.
     */
    SimpleMethodModel(Object object, Method method)
    {
        this.object = object;
        this.method = method;
    }

    /**
     * Invokes the method, passing it the arguments from the list.
     */
    public TemplateModel exec(List arguments)
        throws
        TemplateModelException
    {
        try
        {
            Object[] args = null;
            if(arguments != null)
            {
                int size = arguments.size();
                args = new Object[size];
                Iterator it = arguments.iterator();
                int i = 0;
                while(it.hasNext())
                    args[i++] = BeansWrapper.unwrap(it.next());
                coerceBigDecimals(args);
            }
            
            return BeansWrapper.wrap(method.invoke(object, args));
        }
        catch(Exception e)
        {
            StringWriter sw = new StringWriter();
            PrintWriter pw = new PrintWriter(sw);
            pw.println("Method " + method.getName() + " on " + object.getClass().getName());
            e.printStackTrace(pw);
            throw new TemplateModelException(sw.toString());
        }
    }

    public TemplateModel get(int index)
        throws
        TemplateModelException
    {
        return exec(Collections.singletonList(new SimpleNumber(new Integer(index))));
    }

    public boolean isEmpty()
    {
        return false;
    }

    public int size() throws TemplateModelException
    {
        throw new TemplateModelException("_size is unsupported for: " + getClass().getName());
    }

    /**
     * Converts any {@link BigDecimal}s in the passed array to the type of
     * the corresponding formal argument of the method.
     */
    private void coerceBigDecimals(Object[] args)
    {
        Class[] formalTypes = null;
        for(int i = 0, l = args.length; i < l; ++i)
        {
            Object arg = args[i];
            if(arg instanceof BigDecimal)
            {
                BigDecimal bd = (BigDecimal)arg;
                if(formalTypes == null)
                {
                    formalTypes = method.getParameterTypes();
                    if(formalTypes.length != l)
                    {
                        // This will die anyway due to incorrect number of
                        // arguments, so there's no point in checking
                        return;
                    }
                }
                Class formalType = formalTypes[i];
                // int is expected in most situations, so we check it first
                if(formalType == Integer.TYPE || formalType == Integer.class)
                {
                    args[i] = new Integer(bd.intValue());
                }
                else if(formalType == Double.TYPE || formalType == Double.class)
                {
                    args[i] = new Double(bd.doubleValue());
                }
                else if(formalType == Long.TYPE || formalType == Long.class)
                {
                    args[i] = new Long(bd.longValue());
                }
                else if(formalType == Float.TYPE || formalType == Float.class)
                {
                    args[i] = new Float(bd.floatValue());
                }
                else if(formalType == Short.TYPE || formalType == Short.class)
                {
                    args[i] = new Short(bd.shortValue());
                }
                else if(formalType == Byte.TYPE || formalType == Byte.class)
                {
                    args[i] = new Byte(bd.byteValue());
                }
            }
        }
    }
}
