/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.ext.beans;

import freemarker.ext.util.IdentityHashMap;

import freemarker.template.*;

import java.lang.ref.*;
import java.lang.reflect.*;
import java.util.*;

/**
 * Wraps the static fields and methods of a class in a {@link TemplateHashModel}.
 * Fields are wrapped using {@link BeansWrapper#wrap(Object)}, and
 * methods are wrapped into {@link SimpleMethodModel} instances.
 * Unfortunately, there is currently no support for bean property-style
 * calls of static methods, as in {@link BeanModel}.
 * @author Attila Szegedi, <a href="mailto:szegedia@freemail.hu">szegedia@freemail.hu</a>
 * @version 1.0
 */
public final class StaticModel
implements
	TemplateHashModel
{
	private static final Map modelCache = new IdentityHashMap();
        private static Set cachedClassNames = new HashSet();

	private Class clazz;
	private final Map map = new HashMap();

	private StaticModel(Class clazz)
	throws
		IllegalAccessException
	{
		this.clazz = clazz;
		populate();
	}

	/**
	 * Creates a model representing the static fields and
	 * methods of the passed class. Models are cached, meaning
	 * that calling create twice with a same class parameter
	 * will yield the same object.
	 * @throws IllegalAccessException if the passed class is not public
	 */
	public static final StaticModel create(Class clazz)
	throws
		IllegalAccessException
	{
		StaticModel model = null;
		synchronized(modelCache)
		{
			model = (StaticModel)modelCache.get(clazz);
                        if(model == null)
                        {
                            if(cachedClassNames.contains(clazz.getName()))
                            {
                                // Class reloads detected, throw away caches
                                modelCache.clear();
                                cachedClassNames = new HashSet();
                                BeanModelBase.clearModelCache();
                            }
                            model = new StaticModel(clazz);
                            modelCache.put(clazz, model);
                            cachedClassNames.add(clazz.getName());
                        }
		}
                return model;
	}

	/**
	 * Returns the field or method named by the <tt>key</tt>
	 * parameter.
	 */
	public TemplateModel get(String key)
	throws
		TemplateModelException
	{
		Object model = map.get(key);
		// Simple method, overloaded method or final field -- these have cached 
		// template models
		if(model instanceof TemplateModel)
			return (TemplateModel)model;
		// Non-final field; this must be evaluated on each call.
		if(model instanceof Field)
		{
			try
			{
				return BeansWrapper.wrap(((Field)model).get(null));
			}
			catch(IllegalAccessException e)
			{
				throw new TemplateModelException("Illegal access for field " + key + " of class " + clazz.getName());
			}
		}

		throw new TemplateModelException("No such key: " + key + " in class " + clazz.getName());
	}

	/**
	 * Returns true if there is at least one public static
	 * field or method in the underlying class.
	 */
	public boolean isEmpty()
	{
		return map.isEmpty();
	}

	private void populate()
	throws
		IllegalAccessException
	{
	    if(!Modifier.isPublic(clazz.getModifiers()))
	    {
	        throw new IllegalAccessException("Can't wrap the non-public class " + clazz.getName());
	    }
	    
		Field[] fields = clazz.getFields();
		for(int i = 0; i < fields.length; ++i)
		{
			Field field = fields[i];
			int mod = field.getModifiers();
			if(Modifier.isPublic(mod) && Modifier.isStatic(mod))
			{
				if(Modifier.isFinal(mod))
					try
					{
					    // public static final fields are evaluated once and
					    // stored in the map
						map.put(field.getName(), BeansWrapper.wrap(field.get(null)));
					}
					catch(IllegalAccessException e)
					{
						// Intentionally ignored
					}
				else
					// This is a special flagging value: Field in the map means
					// that this is a non-final field, and it must be evaluated
					// on each get() call.
					map.put(field.getName(), field);
			}
		}
		Method[] methods = clazz.getMethods();
		for(int i = 0; i < methods.length; ++i)
		{
			Method method = methods[i];
			int mod = method.getModifiers();
			if(Modifier.isPublic(mod) && Modifier.isStatic(mod))
			{
			    String name = method.getName();
			    Object obj = map.get(name);
			    if(obj == null)
			    {
			        map.put(name, method);
			    }
			    else if(obj instanceof Method)
			    {
			        MethodMap methodMap = new MethodMap(name);
			        methodMap.addMethod((Method)obj);
			        methodMap.addMethod(method);
			        map.put(name, methodMap);
			    }
			    else
			    {
			        MethodMap methodMap = (MethodMap)obj;
			        methodMap.addMethod(method);
			    }
			}
		}
		for (Iterator entries = map.entrySet().iterator(); entries.hasNext();)
        {
            Map.Entry entry = (Map.Entry) entries.next();
            Object value = entry.getValue();
            if(value instanceof Method)
            {
                entry.setValue(new SimpleMethodModel(null, (Method)value));
            }
            else if(value instanceof MethodMap)
            {
                entry.setValue(new OverloadedMethodModel(null, (MethodMap)value));
            }
        }
	}
}
