/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.ext.servlet;

import freemarker.template.*;

import java.io.*;
import java.util.*;
import java.text.*;

import javax.servlet.*;
import javax.servlet.http.*;

/**
 * <p>This is a general-purpose FreeMarker servlet with following features:</p>
 * <ul>
 * <li>It utilizes a TemplateCache that is fully parameterizable</li>
 * <li>It makes all request, request parameters, session, and servlet
 * context attributes available to templates through "Request", "RequestParameters",
 * "Session", and "Application" variables. The implementation for these variables
 * is efficient; rather than copying the attributes and parameters into {@link freemarker.template.SimpleHash},
 * special wrapper classes are used that implement the {@link freemarker.template.TemplateHashModel}
 * interface.</li>
 * </ul>
 * <p>Supported initialization parameters are:</p>
 * <ul>
 * <li><strong>TemplatePath</strong> specifies the location of the templates.
 * By default, the location is the web application-relative URI of
 * the location.<br>
 * Alternatively, you can prepend it with <tt>file://</tt> to indicate a literal
 * path in the file system (i.e. <tt>file:///var/www/project/templates/</tt>). 
 * Note that three slashes were used to specify an absolute path.<br>
 * Also, you can prepend it with <tt>class://path/to/template/package</tt> to
 * indicate that you want to load templates from the specified package
 * accessible through the classloader of the servlet.<br>
 * Default value is <tt>class://</tt> (that is, the root of the class hierarchy).
 * <i>Note that this default is different than the default in FreeMarker 1.x, when
 * it defaulted <tt>/</tt>, that is to loading from the webapp root directory.</i></li>
 * <li><strong>TemplateUpdateInterval</strong> value for update interval of the
 * template cache in milliseconds. Default is <tt>60000</tt> (ten minutes)</li>
 * <li><strong>NoCache</strong> if set to true, generates headers in the response
 * that advise the HTTP client not to cache the returned page. Default <tt>false</tt></li>
 * </ul>
 * @author Attila Szegedi, <a href="mailto:szegedia@freemail.hu">szegedia@freemail.hu</a>
 * @version 1.0
 */

public class FreemarkerServlet
    extends
    HttpServlet
{
    public static final long serialVersionUID = -2440216393145762479L;

    private static final String INITPARAM_TEMPLATE_PATH = "TemplatePath";
    private static final String INITPARAM_TEMPLATE_DELAY = "TemplateUpdateInterval";
    private static final String INITPARAM_TEMPLATE_DELAY_DEFAULT = "60000";
    private static final String INITPARAM_NOCACHE = "NoCache";
    private static final String INITPARAM_ENCODING = "DefaultEncoding";

    private static final String KEY_REQUEST = "Request";
    private static final String KEY_REQUEST_PARAMETERS = "RequestParameters";
    private static final String KEY_SESSION = "Session";
    private static final String KEY_APPLICATION = "Application";

    // Note these names start with dot, so they're essentially invisible from
    // a freemarker script.
    private static final String ATTR_REQUEST_MODEL = ".freemarker.Request";
    private static final String ATTR_REQUEST_PARAMETERS_MODEL = ".freemarker.RequestParameters";
    private static final String ATTR_SESSION_MODEL = ".freemarker.Session";
    private static final String ATTR_APPLICATION_MODEL = ".freemarker.Application";

    private static final String EXPIRATION_DATE;

    static
    {
        // Generate expiration date that is one year from now in the past
        GregorianCalendar expiration = new GregorianCalendar();
        expiration.roll(Calendar.YEAR, -1);
        SimpleDateFormat httpDate = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss z", java.util.Locale.US);
        EXPIRATION_DATE = httpDate.format(expiration.getTime());
    }

    private TemplateCache cache;
    private String templatePath;
    private String defaultEncoding;
    private boolean nocache;
    protected boolean debug;

    public void init() throws ServletException
    {
        try {
            createCache();
        } catch (IOException ioe) {
           throw new ServletException(ioe);
        }
        // Set browser caching policy
        String strnocache = getServletConfig().getInitParameter(INITPARAM_NOCACHE);
        nocache = ("true".equalsIgnoreCase(strnocache) || "yes".equalsIgnoreCase(strnocache));

        String strdebug = getServletConfig().getInitParameter("debug");
        debug = ("true".equalsIgnoreCase(strdebug) || "yes".equalsIgnoreCase(strdebug));
    }

    protected void createCache() throws IOException
    {
        templatePath = getInitParameter(INITPARAM_TEMPLATE_PATH);
        defaultEncoding = getInitParameter(INITPARAM_ENCODING);
        if(templatePath == null)
            templatePath = "class://";
        FileTemplateCache cache = null;
        if(templatePath.startsWith("class://")) {
            // substring(7) is intentional as we "reuse" the last slash
            cache = new FileTemplateCache(getClass(), templatePath.substring(7));
        }
        else {
            if(templatePath.startsWith("file://")) {
                templatePath = templatePath.substring(7);
            }
            else {
                templatePath = getServletContext().getRealPath(templatePath);
            }
            cache = new FileTemplateCache(new File(templatePath));
        }
        
        if (defaultEncoding != null) {
            cache.setDefaultEncoding(defaultEncoding);
        }
        try {
            String strUpdateInterval = getServletConfig().getInitParameter(INITPARAM_TEMPLATE_DELAY);
            if(strUpdateInterval != null)
                cache.setDelay(Integer.parseInt(strUpdateInterval));
        } catch(NumberFormatException e) {
            // Intentionally ignored
        }
        this.cache = cache;
    }

    public void doGet(HttpServletRequest request, HttpServletResponse response)
    throws
        ServletException,
        IOException
    {
        process(request, response);
    }

    public void doPost(HttpServletRequest request, HttpServletResponse response)
    throws
        ServletException,
        IOException
    {
        process(request, response);
    }

    private void process(HttpServletRequest request, HttpServletResponse response)
    throws
        ServletException,
        IOException
    {
        // Give chance to subclasses to perform preprocessing
        if(preprocessRequest(request, response))
            return;

        String path = request.getPathInfo();
        if(path == null) path = "";

        if(debug)
            System.err.println("Requested template: " + path);

        Template template = cache.getTemplate(path);
        if(template == null)
        {
            response.sendError(HttpServletResponse.SC_NOT_FOUND);
            return;
        }

        // Set cache policy
        setBrowserCachingPolicy(response);

        SimpleHash params = new SimpleHash();

        // Create hash model wrapper for servlet context (the application)
        ServletContext servletContext = getServletContext();
        ServletContextHashModel servletContextModel = (ServletContextHashModel)servletContext.getAttribute(ATTR_APPLICATION_MODEL);
        if(servletContextModel == null)
        {
            servletContextModel = new ServletContextHashModel(servletContext);
            servletContext.setAttribute(ATTR_APPLICATION_MODEL, servletContextModel);
            initializeServletContext(request, response);
        }
        params.put(KEY_APPLICATION, servletContextModel);

        // Create hash model wrapper for session
        HttpSession session = request.getSession();
        HttpSessionHashModel sessionModel = (HttpSessionHashModel)session.getAttribute(ATTR_SESSION_MODEL);
        if(sessionModel == null)
        {
            sessionModel = new HttpSessionHashModel(session);
            session.setAttribute(ATTR_SESSION_MODEL, sessionModel);
            initializeSession(request, response);
        }
        params.put(KEY_SESSION, sessionModel);

        // Create hash model wrapper for request
        HttpRequestHashModel requestModel = (HttpRequestHashModel)request.getAttribute(ATTR_REQUEST_MODEL);
        if(requestModel == null)
        {
            requestModel = new HttpRequestHashModel(request);
            request.setAttribute(ATTR_REQUEST_MODEL, requestModel);
        }
        params.put(KEY_REQUEST, requestModel);

        // Create hash model wrapper for request parameters
        HttpRequestParametersHashModel requestParametersModel = (HttpRequestParametersHashModel)request.getAttribute(ATTR_REQUEST_PARAMETERS_MODEL);
        if(requestParametersModel == null)
        {
            requestParametersModel = createRequestParametersHashModel(request);
            request.setAttribute(ATTR_REQUEST_PARAMETERS_MODEL, requestParametersModel);
        }
        params.put(KEY_REQUEST_PARAMETERS, requestParametersModel);

        // Give subclasses a chance to hook into preprocessing
        if(preTemplateProcess(request, response, template, params))
        {
            try
            {
                // Process the template
                template.process(params, response.getWriter());
            }
            catch( TemplateException te )
            {
                this.log("Error processing request", te);
            }
            finally
            {
                // Give subclasses a chance to hook into postprocessing
                postTemplateProcess(request, response, template, params);
            }
        }
    }

    /**
     * Called as the first step in request processing, before the templating mechanism
     * is put to work. By default does nothing and returns false. This method is
     * tipically overridden to manage serving of non-template resources (i.e. images)
     * that reside in template directory.
     * @param request the HTTP request
     * @param response the HTTP response
     * @return true to indicate this method has processed the request entirely,
     * and that the further request processing should not take place.
     */
    protected boolean preprocessRequest(HttpServletRequest request, HttpServletResponse response)
    throws
        ServletException,
        IOException
    {
        return false;
    }

    protected final String getTemplatePath()
    {
        return templatePath;
    }

    protected HttpRequestParametersHashModel createRequestParametersHashModel(HttpServletRequest request)
    {
        return new HttpRequestParametersHashModel(request);
    }

    /**
     * Called when servlet detects in a request processing that
     * application-global (that is, ServletContext-specific) attributes are not yet
     * set.
     * This is a generic hook you might use in subclasses to perform a specific
     * action on first request in the context. By default it does nothing.
     * @param request the actual HTTP request
     * @param response the actual HTTP response
     */
    protected void initializeServletContext(HttpServletRequest request, HttpServletResponse response)
        throws
        ServletException,
        IOException
    {
    }

    /**
     * Called when servlet detects in a request processing that
     * session-global (that is, HttpSession-specific) attributes are not yet
     * set.
     * This is a generic hook you might use in subclasses to perform a specific
     * action on first request in the session. By default it does nothing.
     * @param request the actual HTTP request
     * @param response the actual HTTP response
     */
    protected void initializeSession(HttpServletRequest request, HttpServletResponse response)
        throws
        ServletException,
        IOException
    {
    }

    /**
     * Called before the execution is passed to template.process().
     * This is a generic hook you might use in subclasses to perform a specific
     * action before the template is processed. By default does nothing.
     * A typical action to perform here is to inject application-specific
     * objects into the <tt>TemplateModelRoot</tt>.
     * @param request the actual HTTP request
     * @param response the actual HTTP response
     * @param template the template that will get executed
     * @param data the data that will be passed to the template
     * @return true to process the template, false to suppress template processing.
     */
    protected boolean preTemplateProcess(HttpServletRequest request, HttpServletResponse response, Template template, TemplateModelRoot data)
        throws
        ServletException,
        IOException
    {
        return true;
    }

    /**
     * Called after the execution returns from template.process().
     * This is a generic hook you might use in subclasses to perform a specific
     * action after the template is processed. It will be invoked even if the
     * template processing throws an exception. By default does nothing.
     * @param request the actual HTTP request
     * @param response the actual HTTP response
     * @param template the template that was executed
     * @param data the data that was passed to the template
     */
    protected void postTemplateProcess(HttpServletRequest request, HttpServletResponse response, Template template, TemplateModelRoot data)
        throws
        ServletException,
        IOException
    {
    }

    /**
     * If the parameter "nocache" was set to true, generate a set of headers
     * that will advise the HTTP client not to cache the returned page.
     */
    private void setBrowserCachingPolicy(HttpServletResponse response)
    {
        if(nocache)
        {
            // HTTP 1.1 browsers should defeat caching on this header
            response.setHeader("Cache-Control", "no-cache");
            // HTTP 1.0 browsers should defeat caching on this header
            response.setHeader("Pragma", "no-cache");
            // Last resort for those that ignore all of the above
            response.setHeader("Expires", EXPIRATION_DATE);
        }
    }
}
