/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project
 * contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template;

import java.util.*;
import java.io.*;
import java.net.URL;
import java.net.URLConnection;

/**
 * A <tt>Cache</tt> that loads templates from the filesystem
 * or via the Class.getResource() mechanism.
 *
 * <p>Templates are loaded into the cache only when requested.
 * The cache maintains a table of when each file was last read 
 * from disk. If more than a certain period of time has elapsed
 * since the last disk read, the cache will check for a newer
 * version. If the file has been deleted, it is removed from the cache
 * and the getTemplate() method will return null. How much time
 * has to elapse before the cache looks for a newer version on disk
 * can be set via the setDelay() method. By default, it is 5 seconds.
 * 
 * <p>If a cache is created with the constructor with no arguments,
 * it is assumed that you wish to load templates relative to the
 * classloader classpath. If you create a cache with a 
 * <code>java.io.File</code> as a parameter, this must be 
 * an existing directory on the file system on which this code
 * has read permission. If you use the constructor that takes a
 * <CODE>Class</CODE> as an argument, this is assumed to
 * be the class that you will use for the <CODE>Class.getResource()</CODE>
 * call. You can also a set a path relative to the ClassLoader loading
 * mechanism.
 * 
 * <p>You can set a default encoding. Also, there is a version of the
 * getTemplate() method that specifies encoding. The method that
 * takes a Locale as an argument deduces the desired encoding from the locale.
 * Note that this method will look for localized versions of the file in 
 * question. For example, myFile_fr_CA.html for a file localized to Canadian
 * French. If that file is not found, it looks for myFile_fr.html, which 
 * would be a file localized to French, without specifying a country, and
 * then, if that file is not found, myFile.html is used.
 *
 * @version $Id: FileTemplateCache.java,v 1.50 2002/04/30 23:18:20 revusky Exp $
 */
public class FileTemplateCache implements TemplateCache {

    // How long should elapse before we check for a newer version
    // of the file.
    private long delay = 5000L; // default of five seconds

    // Our cache of the templates.
    private Map templateCache = Collections.synchronizedMap(new HashMap());

    //A table of the last time we checked for a newer version.
    private Map lastLoadedTable = Collections.synchronizedMap(new HashMap());

    private String path = "/";

    // The listeners to notify when events occur.
    private ArrayList listeners;

    // The class to use for loading via the classpath mechanism.
    private Class clase = FileTemplateCache.class;

    // The explicit location on the file system to look for files
    // This is null if loading is done via the class loader mechanism.
    private File explicitDirectory;

    // The default encoding to use when reading templates.
    private String defaultEncoding = System.getProperty("file.encoding");

    /**
     * This will create an instance that simply attempts
     * to load files via the classloader of this class
     */
    public FileTemplateCache() {
        this.clase = FileTemplateCache.class;
        this.path = "/";
    }

    /**
     * This will create an instance that simply attempts
     * to load files via the classloader mechanism of the
     * specified class loader. Typical usage would be
     * to call this in the init() method of a servlet
     * and the Class passed in would be that servlet class.
     * @param clase the class whose classloader is used
     * for loading the templates
     */
    public FileTemplateCache(Class clase) {
        this.clase = clase;
        this.path = "/";
    }

    /**
     * Constructs a cache which loads template files
     * relative to the classloader of a given class.
     * This is very useful for servlet deployment, since
     * the servlet class can be passed in as an argument.
     * If the path is specified, it represents a path 
     * relative to the Class.getResource() mechanism.
     */
    public FileTemplateCache(Class clase, String path) throws IOException {
        if (path == null || path.equals("")) {
            path = "/";
        }
        if (path.charAt(path.length() -1) != '/') {
            path += "/";
        }
        if (path.charAt(0) != '/') {
            path = "/" + path;
        }
        this.path = path;
        this.clase = clase;
    }

    /**
     * Constructs a cache which loads template files 
     * from an explicit directory on the file system.
     */
    public FileTemplateCache(File explicitDirectory) throws IOException {
        explicitDirectory = explicitDirectory.getCanonicalFile();
        if (!explicitDirectory.exists()) {
            throw new FileNotFoundException("No such directory: " + explicitDirectory);
        }
        if (!explicitDirectory.isDirectory()) {
            throw new FileNotFoundException(explicitDirectory.toString() + " is not a directory.");
        }
        if (!explicitDirectory.canRead()) {
            throw new IOException("Cannot read directory: " + explicitDirectory);
        }
        this.explicitDirectory = explicitDirectory;
    }

    /**
     * Sets the delay in seconds between checking for newer
     * versions of a file.
     * @param delay how long since the last access before checking for a newer version of the file.
     */
    public void setDelay(int delay) {
        this.delay = 1000L * delay;
    }

    public void setDefaultEncoding(String encoding) {
        defaultEncoding = encoding;
    }

    public Template getTemplate(String name, String encoding) throws IOException {
        Template template = (Template) templateCache.get(name);
        Date now = new Date();
        Date lastChecked = (Date) lastLoadedTable.get(name);
        long lastCheck = lastChecked==null ? 0L : lastChecked.getTime();
        long elapsed = now.getTime() - lastCheck;
        if (elapsed < delay) {
            return template;
        }
        lastLoadedTable.put(name, now);
        if (explicitDirectory != null) { // Absolute location on filesystem.
            File f = new File(explicitDirectory, name);
            if (!f.exists()) {
                if (template != null) {
                    template = null;
                    templateCache.remove(name);
                }
            }
            else if (f.lastModified() > lastCheck) {
                InputStream is = new FileInputStream(f);
                template = new Template(this, is, encoding);
                template.setName(name);
                templateCache.put(name, template);
                logFileUpdate(name);
            }
        }
        else { // Get it relative to the classpath.
            URL url = clase.getResource(path + name);
            if (url == null) {
                // REVISIT this.
                url = clase.getResource(path.substring(1) + name);
            }
            if (url == null) {
                if (template != null) {
                    template = null;
                }
            }
            else {
                long lastModified = 0L;
                try {
                    lastModified = new File(url.getFile()).lastModified();
                }
                catch (SecurityException se) {
                }
                if (template == null || lastModified==0 || lastCheck<lastModified) {
                    URLConnection conn = url.openConnection();
                    InputStream is = conn.getInputStream();
                    template = new Template(this, is, encoding);
                    template.setName(name);
                    templateCache.put(name, template);
                    logFileUpdate(name);
                }
            }
        }
        return template;
    }

    /**
     * Will attempt to find a an appropriately named file
     * given the name and the locale. This uses the standard
     * convention of myfile_en_US.html for example, for a 
     * file in US English. Since the encoding is not specified,
     * we deduce/guess it from the locale.
     */
    public Template getTemplate(String name, Locale locale) throws IOException {
        return getTemplate(name, locale, null);
    }

    /**
     * Will attempt to find a an appropriately named file
     * given the name and the locale. This uses the standard
     * convention of myfile_en_US.html for example, for a 
     * file in US English. 
     */
    public Template getTemplate(String name, Locale locale, String encoding)  throws IOException {
        Template result = null;
        if (encoding == null) {
            encoding = LocaleUtil.getCharset(locale);
        }
        String[] filenames = LocaleUtil.getFilenames(name, locale);
        for (int i=0; i<filenames.length; i++) {
            if (filenames[i] != null) {
                if (i == 3) {
                    // If this is not a localized file, go back to 
                    // the default locale and encoding.
                    locale = Locale.getDefault();
                    encoding = defaultEncoding;
                }
                result = getTemplate(filenames[i], encoding);
                if (result != null) {
                    result.setLocale(locale);
                    break;
                }
            }
        }
        return result;
    }

    public Template getTemplate(String name) throws IOException {
        return getTemplate(name, defaultEncoding);
    }

    public void outputRawFile(String filename, String encoding, Writer out) 
        throws IOException 
    {
        InputStream is = null;
        if (explicitDirectory != null) {
            File f = new File(explicitDirectory, filename);
            is = new FileInputStream(f);
        }
        else {
            if (clase != null) {
                is = clase.getResourceAsStream(path + filename);
            }
            if (is == null) {
                is = getClass().getResourceAsStream(filename);
            }
        }
        if (is == null) {
            throw new FileNotFoundException("File : " + filename + " not found.");
        }
        if (encoding == null) {
            encoding = defaultEncoding;
        }
        InputStreamReader isr = new InputStreamReader(is, encoding);
        BufferedReader reader = new BufferedReader(isr, 0x1000);
        while (true) {
            int c = reader.read();
            if (c != -1) {
                out.write(c);
            } 
            else {
                break;
            }
        }
    }

    /**
     * This method does nothing. You can override it
     * in a subclass in order to log when a template cache
     * reloads a file due to it having been modified.
     */
    protected void logFileUpdate(String name) {
    }


    public void clear() {
        lastLoadedTable.clear();
        templateCache.clear();
    }
}
