/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template;

import java.util.*;
import java.text.*;

/**
 * A holder for some locale-related routines.
 */

public class LocaleUtil {
    private static String defaultEncoding = System.getProperty("file.encoding");
    private static HashMap encodingMap = new HashMap();
    private static HashMap decimalSymMap = new HashMap();
    private static final Character DOT = new Character('.');
    private static final Character COMMA = new Character(',');

    static {
        encodingMap.put("ar", "ISO-8859-6");
        encodingMap.put("be", "ISO-8859-5");
        encodingMap.put("bg", "ISO-8859-5");
        encodingMap.put("ca", "ISO-8859-1");
        encodingMap.put("cs", "ISO-8859-2");
        encodingMap.put("da", "ISO-8859-1");
        encodingMap.put("de", "ISO-8859-1");
        encodingMap.put("el", "ISO-8859-7");
        encodingMap.put("en", "ISO-8859-1");
        encodingMap.put("es", "ISO-8859-1");
        encodingMap.put("et", "ISO-8859-1");
        encodingMap.put("fi", "ISO-8859-1");
        encodingMap.put("fr", "ISO-8859-1");
        encodingMap.put("hr", "ISO-8859-2");
        encodingMap.put("hu", "ISO-8859-2");
        encodingMap.put("is", "ISO-8859-1");
        encodingMap.put("it", "ISO-8859-1");
        encodingMap.put("iw", "ISO-8859-8");
        encodingMap.put("ja", "Shift_JIS");
        encodingMap.put("ko", "EUC-KR");    
        encodingMap.put("lt", "ISO-8859-2");
        encodingMap.put("lv", "ISO-8859-2");
        encodingMap.put("mk", "ISO-8859-5");
        encodingMap.put("nl", "ISO-8859-1");
        encodingMap.put("no", "ISO-8859-1");
        encodingMap.put("pl", "ISO-8859-2");
        encodingMap.put("pt", "ISO-8859-1");
        encodingMap.put("ro", "ISO-8859-2");
        encodingMap.put("ru", "ISO-8859-5");
        encodingMap.put("sh", "ISO-8859-5");
        encodingMap.put("sk", "ISO-8859-2");
        encodingMap.put("sl", "ISO-8859-2");
        encodingMap.put("sq", "ISO-8859-2");
        encodingMap.put("sr", "ISO-8859-5");
        encodingMap.put("sv", "ISO-8859-1");
        encodingMap.put("tr", "ISO-8859-9");
        encodingMap.put("uk", "ISO-8859-5");
        encodingMap.put("zh", "GB2312");
        encodingMap.put("zh_TW", "Big5");
    }

    /**
     * Gets the preferred charset for the given locale, 
     * or the default platform encoding if the locale
     * is not in the table.
     *
     * @param loc the locale
     * @return the preferred charset
     */
    static public String getCharset(Locale loc) {
        String charset;

        // Try for a full name match (may include country)
        charset = (String) encodingMap.get(loc.toString());
        if (charset != null) return charset;

        // If a full name didn't match, try just the language
        charset = (String) encodingMap.get(loc.getLanguage());

        // tweaked so it doesn't return null.
        return charset != null ? charset : defaultEncoding;
    }

    /**
     * Returns an array of potential filenames, given 
     * the base filename and the locale.
     * For example, if the filename was foo.html 
     * and the locale was fr_CA,
     * then this method would return [null, "foo_fr_CA.html", "foo_fr.html", "foo.html"]
     * The first element is null, because no variant is defined.
     */

    static public String[] getFilenames(String filename, Locale locale) {
        String extension = "";
        String basename = filename;
        String variant = locale.getVariant();
        String country = locale.getCountry();
        String lang = locale.getLanguage();
        int dotIndex = filename.lastIndexOf('.');
        if (dotIndex > 0) {
            basename = filename.substring(0, dotIndex);
            extension = filename.substring(dotIndex);
        }
        //Level of full specification.
        // We try them in order.
        String filename1 = (variant.length() >0) ?
                           basename + "_"
                           + lang + "_"
                           + country + "_"
                           + variant + extension
                           : null;
        String filename2 = (country.length()  >0) ?
                           basename + "_"
                           + lang + "_"
                           + country + extension
                           : null;
        String filename3 = (lang.length() >0) ?
                           basename + "_"
                           + lang + extension
                           : null;
        return new String[] {
            filename1, filename2, filename3, filename
        };
    }

    /**
     * @param loc the locale
     * @return the decimal separator for a locale, usually either 
     * '.' or ','
     */

    static public char getDecimalSeparator(Locale loc) {
        Character result = (Character) decimalSymMap.get(loc);
        if (result == null) {
            NumberFormat nf = DecimalFormat.getNumberInstance(loc);
            if (nf instanceof DecimalFormat) {
                DecimalFormatSymbols dfs = ((DecimalFormat)nf).getDecimalFormatSymbols();
                char c= dfs.getDecimalSeparator();
                if (c == ',') {
                    decimalSymMap.put(loc, COMMA);
                    return ',';
                }
                else if (c == '.') {
                    decimalSymMap.put(loc, DOT);
                    return '.';
                } 
                else {
                    decimalSymMap.put(loc, new Character(c));
                    return c;
                }
            } else {
                decimalSymMap.put(loc, DOT);
                return '.';
            }
        }
        return result.charValue();
    }

}
