/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template;

import java.util.*;
import java.io.Serializable;

/**
 * <p>A simple implementation of the <tt>TemplateHashModelEx</tt> and
 * <tt>TemplateModelRoot</tt> interfaces, using an underlying
 * <tt>HashMap</tt>.</p>
 *
 * <p><b>Note:</b><br />
 * As of 2.0, this class is unsynchronized by default. To obtain a
 * synchronized wrapper, call the {@link #synchronizedWrapper} method.</p>
 *
 * @version $Id: SimpleHash.java,v 1.14 2002/04/10 14:29:53 revusky Exp $
 * @see SimpleList
 * @see SimpleScalar
 */
public class SimpleHash 
implements TemplateModelRoot, TemplateHashModelEx, Serializable {

    /**
     * @serial the <tt>HashMap</tt> that this <tt>SimpleHash</tt> wraps.
     */
    protected HashMap hash;
    private Locale locale;

    /**
     * Constructs an empty <tt>SimpleHash</tt>.
     */
    public SimpleHash() {
      hash = new HashMap();
    }

    /**
     * Constructs a <tt>SimpleHash</tt> given the backing <tt>Map</tt>.
     *
     * @param hash The Map to use as the backing for this SimpleHash.
     */
    public SimpleHash(Map map) {
        this.hash = new HashMap(map);
    }

    /**
     * Is the underlying hash empty?
     */
    public boolean isEmpty() throws TemplateModelException {
        return hash.isEmpty();
    }

    /**
     * Puts a <tt>TemplateModel</tt> in the hash.
     *
     * @param key the name by which the <tt>TemplateModel</tt> is
     * identified in the template.
     * @param model the <tt>TemplateModel</tt> to store.
     */
    public void put(String key, TemplateModel model) {
        hash.put(key, model);
    }

    /**
     * Puts a string in the hash, by first wrapping the string in a
     * <tt>SimpleScalar</tt>.
     *
     * @param key the name by which the resulting <tt>TemplateModel</tt>
     * is identified in the template.
     * @param s the string to store.
     */
    public void put(String key, String s) {
        put(key, new SimpleScalar(s));
    }

    /**
     * Puts a boolean in the hash, by first wrapping the boolean in a
     * <tt>SimpleScalar</tt>.
     *
     * @param key the name by which the resulting <tt>TemplateModel</tt>
     * is identified in the template.
     * @param b the boolean to store.
     */
    public void put(String key, boolean b) {
        put(key, new SimpleScalar(b));
    }

    public void put(String key, Number num) {
        put(key, new SimpleNumber(num));
    }

    /**
     * Gets a <tt>TemplateModel</tt> from the underlying hash.
     *
     * @param key the name by which the <tt>TemplateModel</tt>
     * is identified in the template.
     * @return the <tt>TemplateModel</tt> referred to by the key,
     * or null if not found.
     */
    public TemplateModel get(String key) throws TemplateModelException {
        return (TemplateModel)hash.get(key);
    }

    /**
     * Removes the given key from the underlying hash.
     *
     * @param key the key to be removed
     */
    public void remove(String key) {
        hash.remove(key);
    }

    /**
     * Convenience method for returning the <tt>String</tt> value of the
     * underlying hash.
     */
    public String toString() {
        return hash.toString();
    }

    public void setLocale(Locale locale) {
        this.locale = locale;
    }

    public Locale getLocale() {
         if (locale == null) {
             locale = Locale.getDefault();
         }
         return locale;
    }

    public int size() throws TemplateModelException {
        return hash.size();
    }

    public TemplateSequenceModel keys() throws TemplateModelException {
        SimpleSequence result = new SimpleSequence();
        for (Iterator it = hash.keySet().iterator(); it.hasNext();) {
            result.add((String) it.next());
        }
        return result;
    }

    public TemplateSequenceModel values() throws TemplateModelException {
        SimpleSequence result = new SimpleSequence();
        for (Iterator it = hash.values().iterator(); it.hasNext();) {
            result.add((String) it.next());
        }
        return result;
    }

    public SimpleHash synchronizedWrapper() {
        return new SynchronizedHash();
    }

    private class SynchronizedHash extends SimpleHash {

        public synchronized boolean isEmpty() throws TemplateModelException {
            return SimpleHash.this.isEmpty();
        }
        
        public synchronized void put(String key, TemplateModel model) {
            SimpleHash.this.put(key, model);
        }

        public synchronized TemplateModel get(String key) throws TemplateModelException {
            return SimpleHash.this.get(key);
        }

        public synchronized void remove(String key) {
            SimpleHash.this.remove(key);
        }

        public synchronized int size() throws TemplateModelException {
            return SimpleHash.this.size();
        }

        public synchronized TemplateSequenceModel keys() throws TemplateModelException {
            return SimpleHash.this.keys();
        }

        public synchronized TemplateSequenceModel values() throws TemplateModelException {
            return SimpleHash.this.values();
        }
    }
}
