/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template;

import java.io.*;
import java.text.NumberFormat;
import java.util.Locale;
import freemarker.template.compiler.Expression;


/**
 * A simple implementation of the <tt>TemplateNumberModel</tt>
 * interface. Note that this class is immutable.
 * @author <A HREF="mailto:jon@revusky.com">Jonathan Revusky</A>
 */
public final class SimpleNumber extends Expression 
implements TemplateNumberModel, Serializable {

    /**
     * @serial the value of this <tt>SimpleNumber</tt> 
     */
    private Number value;
    private NumberFormat format;

    public SimpleNumber(Number value) {
        this.value = value;
    }

    public SimpleNumber(Number value, NumberFormat format) {
        this.value = value;
        this.format = format;
    }

    public SimpleNumber(Number value, Locale locale) {
        this.value = value;
        this.format = NumberFormat.getNumberInstance(locale);
    }

    public SimpleNumber(int val) {
        this.value = new Integer(val);
    }

    public synchronized boolean isEmpty() {
        return value == null;
    }

    public Number getAsNumber() {
        return value;
    }

    /**
     * We use the locale to figure out the 
     * decimal separator to use.
     */
    public String getAsString(Locale loc) {
        if (value == null) {
            return "";
        }
        if (format == null) {
            String val = value.toString();
            if (val.indexOf('.') >=0) {
                char decSep = LocaleUtil.getDecimalSeparator(loc);
                if (decSep != '.') {
                    val = val.replace('.', decSep);
                }
            }
            return val;
        }
        else 
            return format.format(value);
    }

// The methods below implement freemarker.template.expression.Expression
// so that the class can do double duty.

    public String getStringValue(TemplateModelRoot modelRoot) {
        return getAsString(modelRoot.getLocale());
    }

    public Number getNumericalValue(TemplateModelRoot modelRoot) {
        return value;
    }

    public TemplateModel getAsTemplateModel(TemplateModelRoot modelRoot) {
        return this;
    }

    public boolean isTrue(TemplateModelRoot modelRoot) {
        return value != null;
    }

    /**
     * always returns true.
     */
    public boolean isNumerical(TemplateModelRoot modelRoot) {
        return true;
    }

    public String getName() {
        return "the number: '" + value + "'";
    }

    public String toString() {
        return value.toString();
    }
}
