/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template;

import freemarker.template.compiler.Expression;
import freemarker.template.compiler.NonNumericalException;
import java.io.*;
import java.util.Locale;

/**
 * A simple implementation of the <tt>TemplateScalarModel</tt>
 * interface, using a <tt>String</tt> or a <tt>boolean</tt>.
 * As of version 2.0 this object is immutable.
 *
 * @version $Id: SimpleScalar.java,v 1.16 2002/04/26 17:35:18 revusky Exp $
 * @see SimpleSequence
 * @see SimpleHash
 */
public final class SimpleScalar extends Expression 
implements TemplateScalarModel, Serializable {

    public static final SimpleScalar TRUE = new SimpleScalar(true);
    public static final SimpleScalar FALSE = new SimpleScalar(false);

    /**
     * @serial the value of this <tt>SimpleScalar</tt> if it wraps a
     * <tt>String</tt>.
     */
    private String stringValue;

    /**
     * @serial the value of this <tt>SimpleScalar</tt> if it wraps a
     * <tt>boolean</tt>.
     */
    private boolean booleanValue;

    /**
     * @serial true if this <tt>SimpleScalar</tt> wraps a <tt>boolean</tt>
     * value.
     */
    private boolean useBoolean;

    /**
     * Constructs a <tt>SimpleScalar</tt> containing a string value.
     *
     * @param value the string value.
     */
    public SimpleScalar(String value) {
        this.stringValue = value;
    }

    /**
     * Constructs a <tt>SimpleScalar</tt> containing a boolean value.
     *
     * @param value the boolean value.
     */
    public SimpleScalar(boolean value) {
        this.booleanValue = value;
        useBoolean = true;
    }

    public boolean isEmpty() {
        if( useBoolean ) return !booleanValue;
        return ((stringValue == null) || (stringValue.equals("")));
    }

    /**
     * Note that this implementation does not
     * use the locale parameter.
     */
    public String getAsString(Locale locale) {
        if( useBoolean ) {
            return booleanValue ? "true" : ""; 
        }
        else {
            return (stringValue == null) ? "" : stringValue;
        }
    }

    public boolean isTrue(TemplateModelRoot modelRoot) {
        return !isEmpty();
    }

    public String getStringValue(TemplateModelRoot modelRoot) {
        return getAsString(modelRoot.getLocale());
    }

    public Number getNumericalValue(TemplateModelRoot modelRoot) throws TemplateException {
        throw new NonNumericalException();
    }

    /**
     * Always returns false.
     */
    public boolean isNumerical(TemplateModelRoot modelRoot) {
        return false;
    }

    public TemplateModel getAsTemplateModel(TemplateModelRoot modelRoot) {
        return this;
    }

    public String toString() {
        String s = stringValue;
        if (useBoolean) {
            s = booleanValue ? "true" : "";
        }
        return "\"" + s + "\""; 
    }
}
