/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template;

import java.io.Serializable;
import java.util.*;

/**
 * <p>A convenient implementation of a list. This
 * object implements TemplateSequenceModel, using an underlying 
 * <tt>java.util.List</tt> implementation.</p>
 *
 * <p>A <tt>SimpleSequence</tt> can act as a cache for a
 * <tt>TemplateCollectionModel</tt>, e.g. one that gets data from a
 * database.  When passed a <tt>TemplateCollectionModel</tt> as an
 * argument to its constructor, the <tt>SimpleSequence</tt> immediately 
 * copies all the elements and discards the <tt>TemplateCollectionModel</tt>.</p>
 *
 * <p><b>Note:</b><br />
 * As of 2.0, this class is unsynchronized by default. To obtain a
 * synchronized wrapper, call the {@link #synchronizedWrapper} method.</p>
 *
 * @version $Id: SimpleSequence.java,v 1.2 2002/04/16 14:46:00 revusky Exp $
 * @see SimpleHash
 * @see SimpleScalar
 */
public class SimpleSequence implements TemplateSequenceModel, Serializable {

    /**
     * @serial The <tt>List</tt> that this <tt>SimpleSequence</tt> wraps.
     */
    protected List list;

    /**
     * Constructs an empty <tt>SimpleSequence</tt> with an indefinite shelf life.
     */
    public SimpleSequence() {
        list = new ArrayList();
    }

    /**
     * Constructs a <tt>SimpleSequence</tt> from the given <tt>List</tt>
     * It makes a copy for internal use.
     */
    public SimpleSequence( List list ) {
        this.list = new ArrayList(list);
    }

    public SimpleSequence(TemplateCollectionModel tcm) throws TemplateModelException {
        list = new ArrayList();
        for (TemplateModelIterator it = tcm.iterator(); it.hasNext();) {
            list.add(it.next());
        }
    }

    /**
     * Is the underlying <tt>List</tt> empty?
     */
    public boolean isEmpty() throws TemplateModelException {
        return list.isEmpty();
    }

    /**
     * Adds a string to the end of this <tt>SimpleSequence</tt>, by first
     * wrapping the string in a <tt>SimpleScalar</tt>. 
     *
     * @param element the <tt>TemplateModel</tt> to be added.
     */
    public void add(TemplateModel element) {
        list.add(element);
    }

    /**
     * Adds a string to the end of this <tt>SimpleSequence</tt>, by
     * first wrapping the string in a <tt>SimpleScalar</tt>.
     *
     * @param element the string to be added.
     */
    public void add(String s) {
        add(new SimpleScalar(s));
    }

    /**
     * Adds a boolean to the end of this <tt>SimpleSequence</tt>, by first
     * wrapping the boolean in a <tt>SimpleScalar</tt>.  
     *
     * @param element the boolean to be added.
     */
    public void add(boolean b) {
        add(new SimpleScalar(b));
    }
    
    /**
     * @return the specified index in the list
     */
    public TemplateModel get(int i) throws TemplateModelException {
        if(i >= list.size())
            throw new TemplateModelException("IndexOutOfBounds(" + i + ")");
        return (TemplateModel)list.get(i);
    }

    public int size() {
        return list.size();
    }

    /**
     * @return a synchronized wrapper for list.
     */

    public SimpleSequence synchronizedWrapper() {
        return new SynchronizedSequence();
    }

    private class SynchronizedSequence extends SimpleSequence {

        public synchronized boolean isEmpty() throws TemplateModelException {
            return SimpleSequence.this.isEmpty();
         }

        public synchronized void add(TemplateModel element) {
            SimpleSequence.this.add(element);
        }

        public synchronized TemplateModel get(int i) throws TemplateModelException {
            return SimpleSequence.this.get(i);
        }
        
        public synchronized int size() {
            return SimpleSequence.this.size();
        }
    }
}

