/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project
 * contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template;

import java.io.*;
import java.util.*;
import freemarker.template.compiler.*;


/**
 * <p>An application or servlet can instantiate a subclass of <tt>Template</tt>
 * to compile and process an HTML template.
 *
 * <p>You can pass the filename of the template to the constructor, in
 * which case it is compiled immediately.  Once compiled, the template is
 * stored in an an efficient data structure for later use.
 *
 * <p>To process the template and produce HTML, call the
 * <tt>process()</tt> method, which takes a tree of
 * <tt>TemplateModel</tt> objects as its data model.  The root
 * node of the tree must be a <tt>TemplateModelRoot</tt>.
 *
 * <p>Any error messages from exceptions thrown by the data model, or
 * generated by the <tt>Template</tt> during compilation or
 * processing, will be thrown up to the caller to deal with.
 *
 * @version $Id: Template.java,v 1.60 2002/04/26 23:02:42 revusky Exp $
 */

public class Template {

    protected TemplateElement rootElement; // the compiled template
    private Locale locale=Locale.getDefault();
    private Map functions;
    private TemplateCache cache;
    private long lastAccessed;
    private static String defaultEncoding = System.getProperty("file.encoding");
    private static boolean DEBUG_MODE=true;
    public static boolean CLASSIC_COMPATIBILITY = false;
    private String encoding;
    private String name ="";

    /**
     * If the debug mode is set to true (the default currently)
     * all stack traces thrown by the process() method are
     * simply output to the output stream (the web client in 
     * the case of a servlet-based app.) If it is false,
     * the exception is simply rethrown and it is up to
     * the application programmer to catch the exception
     * and deal with it gracefully.
     */
    public static void setDebugMode(boolean b) {
        DEBUG_MODE = b;
    }

    public static boolean getDebugMode() {
        return DEBUG_MODE;
    }

    /**
     * Sets a flag to indicate that we should run in 
     * "classic compatibility mode". In this mode,
     * the == or != comparison between, say, 1 and "1"
     * resolves to true. In the default mode, a comparison
     * of a number with a string throws an exception
     * since it is indicative of a logical flaw in the code.
     * Also, in classic compatibility mode, a single scalar
     * will be treated as a one-item list in an iteration.
     */

    public static void setClassicCompatibility(boolean b) {
        CLASSIC_COMPATIBILITY = b;
    }


    /**
     * Constructs an empty template.
     */
    public Template() {
    }

    /**
     * Constructs a template by compiling it from a file.  Calls
     * <tt>compileFromFile()</tt>.
     *
     * @param filePath the absolute path of the template file
     * to be compiled.
     * @deprecated Expect this to be removed later. 
     * You should either pass in a Reader, or an explicit encoding.
     */
    public Template(String filePath) throws IOException {
        this.name = filePath;
        compileFromFile(filePath);
    }

    /**
     * Constructs a template by compiling it from a file.  Calls
     * <tt>compileFromFile()</tt>.
     *
     * @param file a <tt>File</tt> representing the
     * template file to be compiled.
     * @deprecated Expect this to be removed later. 
     * You should either pass in a Reader, or an explicit encoding.
     */
    public Template(File file) throws IOException {
        this.name = file.toString();
        compileFromFile(file);
    }

    /**
     * Constructs a template by compiling it from an
     * <tt>InputStream</tt>.  Calls
     * <tt>compileFromStream()</tt>.
     *
     * @param stream an <tt>InputStream</tt> from which the
     * template can be read.
     * @deprecated Expect this to be removed later. 
     * You should either pass in a Reader, or an explicit encoding.
     */
    public Template(InputStream stream) throws IOException {
        compileFromStream(stream);
    }

    /**
     * Constructs a template by compiling it from an
     * <tt>Reader</tt>.  Calls <tt>compileFromStream()</tt>.
     *
     * @param stream a <tt>Reader</tt> from which the
     * template can be read.
     */
    public Template(Reader stream) throws IOException {
        compileFromStream(stream);
    }

    public Template(TemplateCache cache, InputStream is) throws IOException {
        this(cache, is, defaultEncoding);
    }

    public Template(TemplateCache cache, InputStream is, String encoding) throws IOException {
        this.cache = cache;
        this.encoding = encoding;
        compileFromStream(new InputStreamReader(is, encoding));
    }

    public Template(TemplateCache cache, InputStream is, Locale locale) throws IOException {
        this.cache = cache;
        if (locale != null) {
            this.locale = locale;
        }
        encoding = LocaleUtil.getCharset(this.locale);
        compileFromStream(new InputStreamReader(is, encoding));
    }

    public Template(TemplateCache cache, Reader r, Locale locale) throws IOException {
        this.cache = cache;
        if (locale != null) {
            this.locale = locale;
        }
        encoding = LocaleUtil.getCharset(this.locale);
        compileFromStream(r);
    }

    public Locale getLocale() {
        return locale;
    }

    /**
     * Sets the cache that this object is stored in.
     */
    public void setCache(TemplateCache cache) {
        this.cache = cache;
    }

    /**
     * Retrieve the cache that this object is stored in.
     */
    public TemplateCache getCache() {
        return this.cache;
    }

    /**
     * Compiles the template from a <tt>Reader</tt>.  If the template has
     * already been compiled, this method does nothing.
     *
     * @param stream an <tt>Reader</tt> from which the
     * template can be read.
     */
    public void compileFromStream(Reader stream) throws IOException {
        if (rootElement != null) {
            return;
        }
        if (!(stream instanceof BufferedReader)) {
            stream = new BufferedReader(stream, 0x1000);
        }
        FMParser parser = new FMParser(this, stream);
        rootElement = parser.Root();
        if (false) {
            // some testing code.
            // dumps the compiled tree out and reparses
            String dump = rootElement.toString();
            StringReader sr = new StringReader(dump);
            parser = new FMParser(this, sr);
            rootElement = parser.Root();
        }
        stream.close();
    }

    /**
     * Reads an unparsed template from a file, by getting the file's
     * <tt>FileInputStream</tt> and using it to call
     * <tt>compileFromStream()</tt>, using the platform's default character
     * encoding.
     *
     * @param file a <tt>File</tt> representing the
     * template file to be compiled.
     * @deprecated Expect this to be removed later. 
     * You should either pass in a Reader, or an explicit encoding.
     */
    public void compileFromFile(File file) throws IOException {
        FileInputStream inputStream;

        if (!file.exists()) {
            throw new FileNotFoundException("Template file " +
                                            file.getName() + " not found.");
        }
        if (!file.canRead()) {
            throw new IOException("Can't read from template file " +
                                  file.getName() + ".");
        }
        inputStream = new FileInputStream(file);
        compileFromStream(inputStream);
    }

    /**
     * Reads the unparsed template from an <tt>InputStream</tt>, using the
     * platform's default character encoding.  If the template has
     * already been compiled, this method does nothing.
     *
     * @param stream an <tt>InputStream</tt> from which the
     * template can be read.
     * @deprecated Expect this to be removed later. 
     * You should either pass in a Reader, or an explicit encoding.
     */
    public void compileFromStream(InputStream stream) throws IOException {
        InputStreamReader streamReader = new InputStreamReader(stream);
        compileFromStream(streamReader);
        streamReader.close();
    }

    /**
     * Compiles the template from an <tt>InputStream</tt>, using the
     * specified character encoding.  If the template has
     * already been compiled, this method does nothing.
     *
     * @param stream an <tt>InputStream</tt> from which the
     * template can be read.
     * @param encoding the character encoding of the input stream
     */
    public void compileFromStream(InputStream stream, String encoding) throws IOException 
    {
        InputStreamReader streamReader = new InputStreamReader(stream, encoding);
        compileFromStream(streamReader);
        streamReader.close();
    }

    /**
     * Reads an unparsed template from a file, by getting the file's
     * <tt>FileInputStream</tt> and using it to call
     * <tt>compileFromStream()</tt>, using the platform's default character
     * encoding.
     *
     * @param filePath the absolute path of the template file
     * to be compiled.
     * @deprecated Expect this to be removed later. 
     * You should either pass in a Reader, or an explicit encoding.
     */
    public void compileFromFile(String filePath)
    throws IOException {

        File file = new File(filePath);
        compileFromFile(file);
    }

    /**
     * Adds a function to the template.
     */
    public synchronized void addFunction(Function function) {
        if (functions == null) {
            functions = new HashMap();
        }
        functions.put(function.getName(), function);
    }

    /**
     * Retrieves a function from the template.  Called by
     * <tt>FunctionCall</tt>s at run-time.
     */
    public Function getFunction(String name) {
        if (functions == null) {
            return null;
        }
        return(Function)functions.get(name);
    }

    /**
     * @return a <tt>Set</tt> of function names (<tt>String</tt> objects)
     * that have been defined for this template.
     */
    public Set getFunctionNames() {
        if (functions == null) {
            return Collections.EMPTY_SET;
        }
        return functions.keySet();
    }

    /**
     * Processes the contents of this <tt>TemplateElement</tt> and
     * outputs the resulting text to a <tt>Writer<tt>.
     *
     * @param modelRoot the root node of the data model.
     * @param out a <tt>Writer</tt> to send the output to.
     * @param dummy A dummy argument that is not used. It is 
     * there as a convenience to keep legacy code compilable.
     */
    public void process(TemplateModelRoot modelRoot, Writer out, Object dummy)
        throws TemplateException, IOException
    {
        process(modelRoot, out);
    }

    /**
     * Processes the template, using data from a template model, and outputs
     * the resulting HTML to a <tt>Writer</tt>. If the DEBUG_MODE
     * is set to true, the stacktrace of any exception thrown 
     * is simply output to the (web) client. Otherwise, the exception
     * is simply rethrown for the caller to handle.
     *
     * @param modelRoot the root node of the data model.  If null, an
     * empty data model is used.
     * @param out a <tt>Writer</tt> to output the HTML to.
     */
    public void process(TemplateModelRoot modelRoot, Writer out) 
        throws TemplateException, IOException 
    {
        if (!DEBUG_MODE) {
            _process(modelRoot, out);
        }
        else {
            try {
                _process(modelRoot, out);
            }
            catch (StopException se) {
                PrintWriter pw = (out instanceof PrintWriter)
                                 ? (PrintWriter) out
                                 : new PrintWriter(out);
                pw.println();
                pw.println("----");
                pw.println("STOP DIRECTIVE ENCOUNTERED HERE");
                pw.println("----\n");
            }
            catch (TemplateException te) {
                PrintWriter pw = (out instanceof PrintWriter)
                                 ? (PrintWriter) out
                                 : new PrintWriter(out);
                te.printStackTrace(pw);
            }
        }
    }

    /**
     * @return the processed template as a string.
     */
    public String toString(TemplateModelRoot root) 
        throws TemplateException, IOException 
    {
        StringWriter sw = new StringWriter();
        _process(root, sw);
        return sw.toString();
    }

    private void _process(TemplateModelRoot root, Writer out) 
        throws TemplateException, IOException 
    {
        if (root == null) {
            root = new SimpleHash();
        }
        root.setLocale(getLocale());
        synchronized(root) {
            rootElement.process(root, out);
        }
    }

    /**
     * Processes the template, using an empty data model, and outputs
     * the resulting HTML to a <tt>Writer</tt>.
     *
     * @param out a <tt>Writer</tt> to output the HTML to.
     */
    public void process(Writer out) throws TemplateException, IOException {
        SimpleHash modelRoot = new SimpleHash();
        process(modelRoot, out);
    }

    /**
     * Processes the template, using data from a template model, and outputs
     * the resulting HTML to an output stream using the encoding
     * for the template.
     *
     * @param modelRoot the root node of the data model.  If null, an
     * empty data model is used.
     * @param out an <tt>OutputStream</tt> to output the HTML to.
     */
    public void process(TemplateModelRoot modelRoot, OutputStream out) 
    throws TemplateException, IOException
    {
        if (!(out instanceof BufferedOutputStream)) {
            out = new BufferedOutputStream(out, 0x1000);
        }
        OutputStreamWriter osw = new OutputStreamWriter(out, getEncoding());
        process(modelRoot, osw);
    }

    /**
     * Dump the raw template in canonical form.
     */
    public void dump(PrintWriter pw) {
        pw.println(rootElement.toString());
    }

    /**
     * Dump the raw template in canonical form.
     */
    public void dump(PrintStream ps) {
        ps.print(rootElement);
    }

    public String getEncoding() {
        if (encoding == null) {
            encoding = LocaleUtil.getCharset(getLocale());
        }
        return encoding;
    }

    public void setEncoding(String encoding) {
        this.encoding = encoding;
    }

    public void setLocale(Locale locale) {
        this.locale = locale;
    }

    public void setName(String name) {
        this.name = name;
    }

    public String getName() {
        return name;
    }
}
