/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project
 * contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template;

import java.util.Locale;
import java.io.Writer;
import java.io.IOException;

/**
 * An interface for caches that store compiled templates. Note that
 * a <CODE>Template</CODE> must be associated with an instance of 
 * <CODE>TemplateCache</CODE> for the &lt;include...&gt; mechanism 
 * to work.
 *
 * @see freemarker.template.FileTemplateCache
 * @version $Id: TemplateCache.java,v 1.17 2002/04/30 13:29:23 revusky Exp $
 */
public interface TemplateCache {

    /**
     * Retrieves a template from the cache. This method assumes
     * defaults for the locale and character encoding.
     *
     * @param name a string uniquely identifying the template.
     * @return the template, or null if not found.
     */
     Template getTemplate(String name) throws IOException;

     /**
      * Retrieves a template corresponding to the name and locale.
      * This method attempts to find a locale-specific version 
      * of the file based on the standard naming scheme, e.g.
      * filename_en_US.html. The encoding is deduced from the locale.
      */
     Template getTemplate(String name, Locale locale) throws IOException;

     /**
      * retrieves a file of the given name with the given encoding.
      */
     Template getTemplate(String name, String encoding) throws IOException;

     /**
      * Retrieves a template corresponding to the name and locale.
      * This method attempts to find a locale-specific version 
      * of the file based on the standard naming scheme, e.g.
      * filename_fr_CA.html. 
      */
     Template getTemplate(String name, Locale locale, String encoding) throws IOException;

     /**
      * Simply outputs the given file from the cache 
      * without any template processing. 
      * If the encoding is null, it uses a default encoding.
      */
     void outputRawFile(String filename, String encoding, Writer out) throws IOException;


     /**
      * Clears the cache.
      */
     void clear();
}
