/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template.compiler;

import freemarker.template.*;
import java.util.Locale;

/**
 * An operator for arithmetic operations. Note that the + operator
 * also does string concatenation for backward compatiability.
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
final class ArithmeticExpression extends Expression {

    static final int ADDITION=1;
    static final int SUBTRACTION=2;
    static final int MULTIPLICATION=3;
    static final int DIVISION=4;
    static final int MODULUS = 5;

    private Expression left, right;
    private int operation;

    ArithmeticExpression(Expression left, Expression right, int operation) {
        this.left = left;
        this.right = right;
        this.operation = operation;
    }

    protected boolean isTrue(TemplateModelRoot modelRoot) throws TemplateException {
        String value = getStringValue(modelRoot);
        return ((value != null) && (!value.equals("")));
    }

    protected String getStringValue(TemplateModelRoot modelRoot) throws TemplateException {
        TemplateModel tm = getAsTemplateModel(modelRoot);
        if (tm == null) {
            return null;
        }
        if (!(tm instanceof TemplateScalarModel)) {
            String msg = "Error " + getLocation()
                        +"\nExpecting scalar here";
            throw new TemplateException(msg);
        }
        return ((TemplateScalarModel) tm).getAsString(modelRoot.getLocale());
    }

    protected TemplateModel getAsTemplateModel(TemplateModelRoot modelRoot)
            throws TemplateException 
    {
        TemplateModel leftModel = left.getAsTemplateModel(modelRoot);
        TemplateModel rightModel = right.getAsTemplateModel(modelRoot);
        if (left == null) {
            String msg = "Error " + getLocation()
                        + "\nExpression " + left + " is undefined.";
            throw new TemplateException(msg);
        }
        if (right == null) {
            String msg = "Error " + getLocation()
                        +"\nExpression " + right + " is undefined.";
            throw new TemplateException(msg);
        }
        boolean isNumerical = (leftModel instanceof TemplateNumberModel)
                              && (rightModel instanceof TemplateNumberModel);
        if (operation == ADDITION && !isNumerical) {
            return concatenate(modelRoot);
        } else {
            Number first = ((TemplateNumberModel) leftModel).getAsNumber();
            Number second = ((TemplateNumberModel) rightModel).getAsNumber();
            Number result = null;
            switch (operation) {
                case ADDITION : 
                    result = NumberUtil.add(first, second);
                    break;
                case SUBTRACTION : 
                    result = NumberUtil.subtract(first, second);
                    break;
                case MULTIPLICATION :
                    result = NumberUtil.multiply(first, second);
                    break;
                case DIVISION :
                    result = NumberUtil.divide(first, second);
                    break;
                case MODULUS :
                    result = NumberUtil.modulus(first, second);
                    break;
                default:
                    throw new TemplateException("unknown operation : " + operation);
            }
            return new SimpleNumber(result);
        }
    }

    private TemplateModel concatenate(TemplateModelRoot root) throws TemplateException 
    {
        Locale locale = root.getLocale();
        TemplateModel tmLeft = left.getAsTemplateModel(root);
        TemplateModel tmRight = right.getAsTemplateModel(root);
        if (tmLeft == null) {
            return tmRight;
        }
        if (tmRight == null) {
            return tmLeft;
        }
        if ((tmLeft instanceof TemplateSequenceModel) && (tmRight instanceof TemplateSequenceModel)){
            TemplateSequenceModel lhs = (TemplateSequenceModel) tmLeft;
            TemplateSequenceModel rhs = (TemplateSequenceModel) tmRight;
            SimpleSequence result = new SimpleSequence();
            for (int i = 0; i<lhs.size(); i++) {
                result.add(lhs.get(i));
            }
            for (int i=0; i<rhs.size(); i++) {
                result.add(rhs.get(i));
            }
            return result;
        } 
        else if ((tmLeft instanceof TemplateScalarModel) && (tmRight instanceof TemplateScalarModel)){
            String s1 = ((TemplateScalarModel) tmLeft).getAsString(locale);
            String s2 = ((TemplateScalarModel) tmRight).getAsString(locale);
            if (s1 == null) {
                return new SimpleScalar(s2);
            }
            if (s2 == null) {
                return new SimpleScalar(s1);
            }
            return new SimpleScalar(s1 + s2);
        }
        else {
            String msg = "Error in the '+' operation "
                         + getLocation()
                         + "\nCannot concatenate these two operands.";
            throw new TemplateException(msg);
        }
    }

    public String toString() {
        String op = " + ";
        if (operation == SUBTRACTION) {
            op = " - ";
        }
        else if (operation == MULTIPLICATION) {
            op = " * ";
        }
        else if (operation == DIVISION) {
            op = " / ";
        }
        else if (operation == MODULUS) {
            op = " % ";
        }
        return left.toString() + op + right.toString();
    }
}
