/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template.compiler;

import freemarker.template.*;
import java.util.*;


/**
 * The ? operator used to get the 
 * functionality of built-in unary operators 
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
final class BuiltIn extends Expression {
    private Expression target;
    private String key;

    static final HashSet keys = new HashSet();

    static {
        // These are the only ones we have now.
        // We throw a parse exception if it's not one of these.
        keys.add("size");
        keys.add("first");
        keys.add("last");
        keys.add("length");
        keys.add("byte");
        keys.add("short");
        keys.add("int");
        keys.add("long");
        keys.add("float");
        keys.add("double");
        keys.add("capitalize");
        keys.add("upper_case");
        keys.add("lower_case");
        keys.add("number");
        keys.add("string");
        keys.add("word_list");
        keys.add("reverse");
        keys.add("keys");
        keys.add("values");
        keys.add("trim");
        keys.add("web_safe");
    }


    BuiltIn(Expression target, String key) throws ParseException {
        this.target = target;
        if (!keys.contains(key)) {
            StringBuffer buf = new StringBuffer("Expecting one of: ");
            for (Iterator it= keys.iterator(); it.hasNext();) {
                if (it.hasNext()) {
                    buf.append(" ");
                } else {
                    buf.append( " or ");
                }
                buf.append(it.next());
                if (it.hasNext()) {
                    buf.append(",\n");
                } 
            }
            throw new ParseException(buf.toString());
        }
        this.key = key.intern();
    }

    protected TemplateModel getAsTemplateModel(TemplateModelRoot modelRoot)
            throws TemplateException {
        Locale locale = modelRoot.getLocale();
        TemplateModel model = target.getAsTemplateModel(modelRoot);
        if (model == null) {
            // SHOULD WE REALLY BE SILENT HERE? REVISIT (JR)
            return null;
        }
        TemplateModel result = null;
        if (model instanceof TemplateHashModelEx) {
            TemplateHashModelEx hash = (TemplateHashModelEx) model;
            if (key=="size") {
                result = new SimpleNumber(hash.size());
            }
            else if (key=="keys") {
                result = hash.keys();
            }
            else if (key=="values") {
                result = hash.values();
            }
        }
        else if (model instanceof TemplateSequenceModel) {
            TemplateSequenceModel sequence = (TemplateSequenceModel) model;
            if (key == "size") {
                result = new SimpleNumber(sequence.size());
            }
            else if (key == "first") {
                result = sequence.get(0);
            }
            else if (key == "last") {
                result = sequence.get(sequence.size() -1);
            }
            else if (key == "reverse") {
                SimpleSequence reversed = new SimpleSequence();
                for (int i=sequence.size()-1; i>=0; i--) {
                    reversed.add(sequence.get(i));
                }
                result = reversed;
            }
        }
        else if (model instanceof TemplateNumberModel) {
            Number num = ((TemplateNumberModel) model).getAsNumber();
            if (key == "string") {
                result = new SimpleScalar(num.toString());
            } else {
                try {
                    result = new SimpleNumber(convertNumber(num, key));
                } catch (ParseException pe) {
                    throw new TemplateException(pe.getMessage() + getLocation());
                }
            }
        }
        else if (model instanceof TemplateScalarModel) {
            String val = ((TemplateScalarModel) model).getAsString(locale);
            if (key == "length") {
                result = new SimpleNumber(val.length());
            }
            else if (key == "lower_case") {
                result = new SimpleScalar(val.toLowerCase(locale));
            }
            else if (key == "upper_case") {
                result = new SimpleScalar(val.toUpperCase(locale));
            }
            else if (key == "capitalize") {
                result = new SimpleScalar(StringUtil.capitalize(val));
            }
            else if (key == "number") {
                try {
                    result = new SimpleNumber(new java.math.BigDecimal(val));
                } catch (RuntimeException re) {
                    String msg = "Error " + getLocation();
                    throw new TemplateException(msg, re);
                }
            }
            else if (key == "string") {
                return model;
            }
            else if (key == "trim") {
                return new SimpleScalar(val.trim());
            }
            else if (key == "word_list") {
                result = new SimpleSequence();
                StringTokenizer st = new StringTokenizer(val);
                while (st.hasMoreTokens()) {
                    ((SimpleSequence) result).add(st.nextToken());
                }
            }
            else if (key == "web_safe") {
                result = new SimpleScalar(StringUtil.convertHTMLEntities(val));
            }
        }
        return result;
    }

    public String toString() {
        return "" + target + "?" + key;
    }

    // Note that this method expects the key to be interned.

    static Number convertNumber(Number num, String key) throws ParseException {
        if (key == "int") {
            return new Integer(num.intValue());
        }
        else if (key == "long") {
            return new Long(num.longValue());
        }
        else if (key == "float") {
            return new Float(num.floatValue());
        }
        else if (key == "double") {
            return new Double(num.doubleValue());
        }
        else if (key == "short") {
            return new Short(num.shortValue());
        }
        else if (key == "byte") {
            return new Byte(num.byteValue());
        }
        else if (key == "number") {
            return num;
        }
        throw new ParseException("expecting one of string, int, long, float, double, short, byte");
    }
}