/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template.compiler;

import freemarker.template.*;

/**
 * A class that handles comparisons.
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */

final class ComparisonExpression extends BooleanExpression {

    static final int EQUALS=1;
    static final int NOT_EQUALS=2;
    static final int LESS_THAN=3;
    static final int GREATER_THAN=4;
    static final int LESS_THAN_EQUALS=5;
    static final int GREATER_THAN_EQUALS=6;

    private Expression left, right;
    private int operation;
    private String opString;

    ComparisonExpression(Expression left, Expression right, String opString) {
        this.left = left;
        this.right = right;
        opString = opString.intern();
        this.opString = opString;
        if (opString == "==" || opString == "=") {
            operation = EQUALS;
        }
        else if (opString == "!=") {
            operation = NOT_EQUALS;
        }
        else if (opString == "\\gt" || opString == ">" || opString == "&gt;") {
            operation = GREATER_THAN;
        }
        else if (opString == "\\gte" || opString == ">=" || opString == "&gt;=") {
            operation = GREATER_THAN_EQUALS;
        }
        else if (opString == "\\lt" || opString == "<" || opString == "&lt;") {
            operation = LESS_THAN;
        }
        else if (opString == "\\lte" || opString == "<=" || opString == "&lt;=") {
            operation = LESS_THAN_EQUALS;
        }
    }

    protected boolean isTrue(TemplateModelRoot modelRoot) throws TemplateException {
        TemplateModel leftModel = left.getAsTemplateModel(modelRoot);
        TemplateModel rightModel = right.getAsTemplateModel(modelRoot);
        boolean leftIsNumerical = (leftModel instanceof TemplateNumberModel);
        boolean rightIsNumerical = (rightModel instanceof TemplateNumberModel);
        if (operation == EQUALS || operation == NOT_EQUALS) {
            return this.checkEquality(modelRoot);
        }
        if (!(leftIsNumerical && rightIsNumerical)) {
            String msg = "Expecting numerical expressions in " 
                         + opString + " comparison " + getLocation();
            if (!leftIsNumerical && !rightIsNumerical) {
                msg += "\nNeither the left nor right hand sides are numerical.";
            }
            else if (!leftIsNumerical) {
                msg += "\nThe left-hand-side " + left + " is not numerical.";
            }
            else if (!rightIsNumerical) {
                msg += "\nThe right-hand-side " + right + " is not numerical.";
            }
            throw new NonNumericalException(msg);
        }
        Number first = ((TemplateNumberModel)leftModel).getAsNumber();
        Number second = ((TemplateNumberModel)rightModel).getAsNumber();
        int comp = NumberUtil.compareNumbers(first, second);
        switch (operation) {
            case LESS_THAN : 
                return comp < 0;
            case GREATER_THAN : 
                return comp > 0;
            case LESS_THAN_EQUALS :
                return comp <=0;
            case GREATER_THAN_EQUALS :
                return comp >=0;
            default :
                throw new TemplateException("unknown operation");
        }
    }

    private boolean checkEquality(TemplateModelRoot modelRoot) throws TemplateException {
        boolean result;
        TemplateModel ltm = left.getAsTemplateModel(modelRoot);
        TemplateModel rtm = right.getAsTemplateModel(modelRoot);
        if (ltm == rtm) {
            return (operation == EQUALS);
        } 
        else if (ltm == null || rtm == null) {
            return operation != EQUALS;
        }
        boolean leftIsNumerical = (ltm instanceof TemplateNumberModel);
        boolean rightIsNumerical = (rtm instanceof TemplateNumberModel);
        if (leftIsNumerical != rightIsNumerical) {
            if (!Template.CLASSIC_COMPATIBILITY) {
                String msg = "Error in " + opString + " comparison " + getLocation();
                if (leftIsNumerical) {
                    msg += ("\nThe left-hand-side " + left + " is numerical " 
                            + "but the right-hand-side " + right + " is not.");
                }
                else {
                    msg += ("\nThe right-hand-side " + right + " is numerical " 
                            + "but the left-hand-side " + left + " is not.");
                }
                throw new NonNumericalException(msg);
            }
        }
        String leftValue = left.getStringValue(modelRoot);
        String rightValue = right.getStringValue(modelRoot);
        if (leftValue == null) {
            result = ((rightValue == null) || (rightValue.equals("")));
        } else if (rightValue == null) {
            result = leftValue.equals("");
        } else {
            result = leftValue.equals(rightValue);
        }
        return operation==EQUALS ? result : !result;
    }

    public String toString() {
        return left.toString() + opString + right.toString();
    }
}
