/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template.compiler;

import freemarker.template.*;
import java.util.ArrayList;
import java.util.Locale;

/**
 * A unary operator that uses the string value of an expression as a hash key.
 * It associates with the <tt>Identifier</tt> or <tt>Dot</tt> to its left.
 */
final class DynamicKeyName extends Expression {

    private Expression nameExpression;
    private Expression target;

    DynamicKeyName(Expression target, Expression nameExpression) {
        this.target = target; 
        this.nameExpression = nameExpression;
    }

    protected TemplateModel getAsTemplateModel(TemplateModelRoot modelRoot) throws TemplateException {
        Locale locale = modelRoot.getLocale();
        TemplateModel targetModel = target.getAsTemplateModel(modelRoot);
        TemplateModel keyModel = nameExpression.getAsTemplateModel(modelRoot);
        if (targetModel == null) {
            // Should we really ignore this? REVISIT (JR)
            return null;
        }
        if (keyModel instanceof TemplateNumberModel) {
            TemplateNumberModel tnm = (TemplateNumberModel) keyModel;
            int index = tnm.getAsNumber().intValue();
            if (targetModel instanceof TemplateSequenceModel) {
                return ((TemplateSequenceModel) targetModel).get(index);
            } else if (targetModel instanceof TemplateScalarModel) {
                // Do we accept a number here? REVISIT (JR)
                String s = ((TemplateScalarModel) targetModel).getAsString(locale);
                try {
                   return new SimpleScalar(s.substring(index, index+1));
                } catch (RuntimeException re) {
                    String msg = "Error " + getLocation();
                    throw new TemplateException(msg, re);
                }
            }
            String mess = "Error " + getLocation()
                         + "\nExpecting TemplateSequenceModel here"
                         + " but expression " + targetModel 
                         + "is a " + targetModel.getClass().getName();
            throw new TemplateException(mess);
        }
        else if (nameExpression instanceof Range) {
            Range range = (Range) nameExpression;
            TemplateModel left = range.left.getAsTemplateModel(modelRoot);
            TemplateModel right = range.right.getAsTemplateModel(modelRoot);
            if (!(left instanceof TemplateNumberModel)) {
                String msg = "Error " + range.getLocation()
                            + "\nExpecting number on left-hand-side of range "
                            + range;
                throw new TemplateException(msg);
            }
            if (!(right instanceof TemplateNumberModel)) {
                String msg = "Error " + range.getLocation()
                            + "\nExpecting number on right-hand-side of range: "
                            + range;
                throw new TemplateException(msg);
            }
            int start = ((TemplateNumberModel) left).getAsNumber().intValue();
            int end = ((TemplateNumberModel) right).getAsNumber().intValue();
            if (targetModel instanceof TemplateSequenceModel) {
                TemplateSequenceModel sequence = (TemplateSequenceModel) targetModel;
                ArrayList list = new ArrayList(1+Math.abs(start-end));
                if (start>end) {
                    for (int i = start; i>=end; i--) {
                        list.add(sequence.get(i));
                    }
                }
                else {
                    for (int i = start; i<=end; i++) {
                        list.add(sequence.get(i));
                    }
                }
                return new SimpleSequence(list);
            } 
            else if (targetModel instanceof TemplateScalarModel) {
                String s = ((TemplateScalarModel) targetModel).getAsString(locale);
                try {
                    return new SimpleScalar(s.substring(start, end+1));
                } catch (RuntimeException re) {
                    String msg = "Error " + getLocation();
                    throw new TemplateException(msg, re);
                }
            }
            String mess = "Error " + getLocation()
                         + "\nExpecting TemplateSequenceModel or TemplateScalarModel here"
                         + " but expression " + targetModel 
                         + "is a " + targetModel.getClass().getName();
            throw new TemplateException(mess);
        }
        else {
            if (!(targetModel instanceof TemplateHashModel)) {
                String msg = "Error " + getLocation()
                             +"\nExpecting a TemplateHashModel here"
                             +" but " + target + " is a " 
                             + targetModel.getClass().getName() + ".";

                throw new TemplateException(msg);
            }
            String key = nameExpression.getStringValue(modelRoot);
            return((TemplateHashModel) targetModel).get(key);
        }
    }

    public String toString() {
        return target.toString() + "[" + nameExpression + "]";
    }
}
