/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template.compiler;

import freemarker.template.*;
import java.io.Writer;
import java.io.IOException;
import java.util.*;

/**
 * An instruction representing a function call.
 */
final class FunctionCall extends TemplateElement {
    private Template template;
    private String functionName;
    private List arguments; // Expression objects.

    FunctionCall(Template template, String functionName, List arguments) {
        this.template = template;
        this.functionName = functionName;
        this.arguments = arguments;
    }

    public void process(TemplateModelRoot modelRoot, Writer out)
        throws TemplateException, IOException
    {
        // Look for the function first in the template, then in the
        // data model, where it might have been put by an Include.
        Function function = template.getFunction(functionName);
        if (function == null) {
            TemplateModel model = modelRoot.get(functionName);
            if (model instanceof FunctionModel) {
                function = ((FunctionModel)model).getFunction();
            }
        }
        if (function == null) {
            String msg = "Expecting function " + getLocation()
                         + "\nfunction " + functionName + " not found.";
            throw new TemplateException(msg);
        }
        List argumentNames = function.getArgumentNames();
        if (argumentNames.size() != arguments.size()) {
            String msg = "Error calling function " + functionName 
                        + "\n" + getLocation()
                        + "\nFound " + arguments.size() + " arguments, "
                        + "expecting " + argumentNames.size() + " arguments.";
            throw new TemplateException(msg);
        }
        // Put the arguments in the data model under the argument names,
        // temporarily hiding the current values of those names.
        Map oldValues = new HashMap();
        Iterator argNamesIterator = argumentNames.iterator();
        Iterator argIterator = arguments.iterator();
        while (argIterator.hasNext()) {
            String argName = (String)argNamesIterator.next();

            // Hide the current value of whatever has this name.
            TemplateModel oldReferent = modelRoot.get(argName);
            if (oldReferent != null) {
                oldValues.put(argName, oldReferent);
            }
            // Put the argument's value in the data model.
            Expression argExpression = (Expression)argIterator.next();
            TemplateModel argModel = argExpression.getAsTemplateModel(modelRoot);
            modelRoot.put(argName, argModel);
        }

        // Call the function.
        try {
            function.process(modelRoot, out);
        } 
        catch (BreakInstruction.Return re) {}
        finally {
            // Restore the old values of the argument names.
            argNamesIterator = argumentNames.iterator();
            while (argNamesIterator.hasNext()) {
                String argName = (String)argNamesIterator.next();
                TemplateModel oldReferent = (TemplateModel)oldValues.get(argName);
                if (oldReferent != null) {
                    modelRoot.put(argName, oldReferent);
                } else {
                    modelRoot.remove(argName);
                }
            }
        }
    }

    public String toString() {
        StringBuffer buf = new StringBuffer("<call ");
        buf.append(functionName);
        buf.append("(");
        int size = arguments.size();
        for (int i=0; i<size; i++) {
            buf.append(arguments.get(i));
            if (i != size-1) {
                buf.append(",");
            }
        }
        buf.append(")>");
        return buf.toString();
    }
}
