/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template.compiler;

import freemarker.template.*;
import java.io.*;
import java.util.Set;
import java.util.Iterator;

/**
 * An instruction that gets another template from a <tt>TemplateCache</tt>,
 * and processes it within the current template.
 */
final class Include extends TemplateElement {

    private Template template;
    private Expression templateName;
    private String encoding;
    private boolean parse;

    /**
     * @param variable the template that this <tt>Include</tt> is a part of.
     * @param templateName the name, in the <tt>TemplateCache</tt>, of the
     * template to be included.
     * @param encoding the encoding to be used or null, if it is a default.
     * @param parsed whether the template should be parsed (or is raw)
     */
    Include(Template template, 
            Expression templateName, 
            String encoding, 
            boolean parse) 
    {
        this.template = template;
        this.templateName = templateName;
        this.encoding = encoding;
        this.parse = parse;
    }

    public void process(TemplateModelRoot modelRoot, Writer out) 
        throws TemplateException, IOException
    {
        TemplateCache cache = template.getCache();
        if (cache == null) {
            String msg = "Error " + getLocation()
                       + "\nThis template is not associated with a template cache, so you cannot do an <include>";
            throw new TemplateException(msg);
        }
        String templateNameString = templateName.getStringValue(modelRoot);
        if( templateNameString == null ) {
            String msg = "Error " + getLocation()
                        + "The expression " + templateName + " is undefined.";
            throw new TemplateException(msg);
        }
        String enc = this.encoding;
        if (enc == null) {
            enc = template.getEncoding();
        }
        if (parse) {
            Template includedTemplate = null;
            try {
                includedTemplate = cache.getTemplate(templateNameString,
                                                              modelRoot.getLocale(),
                                                              enc);
            } 
            catch (ParseException pe) {
                String msg = "Error " + getLocation()
                         + "\nError parsing included template " + templateNameString;
                throw new TemplateException(msg, pe);
            }
            catch (IOException ioe) {
                String msg = "Error " + getLocation()
                         + "\nError reading included template " + templateNameString;
                throw new TemplateException(msg, ioe);
            }
            if (includedTemplate == null) {
                String msg = "Error " + getLocation()
                             + "\nError reading included template " + templateNameString;
                throw new TemplateException("Template \"" +
                        templateNameString + "\" not found in cache.");
            }
            copyFunctions(template, modelRoot);
            copyFunctions(includedTemplate, modelRoot);
            includedTemplate.process(modelRoot, out);
        }
        else {
            try {
                cache.outputRawFile(templateNameString, enc, out); 
            } catch (Exception e) {
                String msg = "Error " + getLocation()
                            + "\nCould not read file: " + templateNameString;
                throw new TemplateException(msg, e);
            }
        }
    }

    /**
     * Copies functions from a template into a data model.
     */
    private void copyFunctions(Template template, TemplateModelRoot modelRoot) {
        Set functionNames = template.getFunctionNames();
        Iterator iterator = functionNames.iterator();

        while (iterator.hasNext()) {
            String functionName = (String)iterator.next();
            Function function = template.getFunction(functionName);
            modelRoot.put(functionName, new FunctionModel(function));
        }
    }

    public String toString() {
        StringBuffer buf = new StringBuffer("<include ");
        buf.append(templateName);
        if (encoding != null) {
            buf.append(" ;");
            buf.append(" encoding=\"");
            buf.append(encoding);
            buf.append("\"");
        }
        if (!parse) {
            if (encoding == null) {
                buf.append(" ;");
            }
            buf.append(" parse=\"n\"");
        }
        buf.append(">");
        return buf.toString();
    }
}
