/*
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template.compiler;

import freemarker.template.*;
import java.io.Writer;
import java.io.IOException;

/**
 * An instruction that processes a list or foreach block
 */
final class IteratorBlock extends TemplateElement {

    private Expression listExpression;
    private TemplateElement block;
    private String indexName, counterName, hasNextName;
    private boolean isForEach;

    /**
     * @param listVariable a variable referring to a TemplateListModel.
     * @param indexName an arbitrary index variable name
     * @param block the block to iterate over
     */
    IteratorBlock(Expression listExpression,
                  String indexName,
                  TemplateElement block,
                  boolean isForEach) 
    {
        this.listExpression = listExpression;
        this.indexName = indexName;
        this.isForEach = isForEach;
        counterName = indexName + "_index";
        hasNextName = indexName + "_has_next";
        this.block = block;
    }

    public void process(TemplateModelRoot modelRoot, Writer out)
        throws TemplateException, IOException
    {
        TemplateModel prevCount = modelRoot.get(counterName);
        TemplateModel prevHasNext = modelRoot.get(hasNextName);
        int indexCount = 0;
        TemplateModel baseModel = listExpression.getAsTemplateModel(modelRoot);
        if (baseModel == null) {
            // SHOULD WE REALLY BE SILENT HERE? REVISIT. (JR)
            return;
        }

        // Hide the old value of the loop index variable.
        TemplateModel oldIndexReferent = modelRoot.get(indexName);
        try {
            if (baseModel instanceof TemplateCollectionModel) {
                TemplateCollectionModel baseListModel = (TemplateCollectionModel)baseModel;
                for (TemplateModelIterator it = baseListModel.iterator(); it.hasNext(); indexCount++) {
                    TemplateModel currElement = it.next();
                    SimpleScalar hasNext = it.hasNext() ? SimpleScalar.TRUE : SimpleScalar.FALSE;
                    SimpleNumber count = new SimpleNumber(indexCount);
                    modelRoot.put(hasNextName, hasNext);
                    modelRoot.put(counterName, count);
                    modelRoot.put(indexName, currElement);
                    block.process(modelRoot, out);
                }
            }
            else if (baseModel instanceof TemplateSequenceModel) {
                TemplateSequenceModel tsm = (TemplateSequenceModel) baseModel;
                int size = tsm.size();
                for (indexCount = 0; indexCount<size; indexCount++) {
                    TemplateModel currElement = tsm.get(indexCount);
                    SimpleScalar hasNext = (indexCount<(size-1)) ? SimpleScalar.TRUE : SimpleScalar.FALSE;
                    SimpleNumber count = new SimpleNumber(indexCount);
                    modelRoot.put(hasNextName, hasNext);
                    modelRoot.put(counterName, count);
                    modelRoot.put(indexName, currElement);
                    block.process(modelRoot, out);
                }
            }
            else if (baseModel instanceof TemplateListModel) {
                // This block exists for backward compatibility
                TemplateListModel baseListModel = (TemplateListModel)baseModel;

                // Loop, setting the index to refer to each element in turn.
                if (!baseListModel.isRewound()) {
                    baseListModel.rewind();
                }
                while (baseListModel.hasNext()) {
                    TemplateModel currElement = baseListModel.next();
                    SimpleScalar hasNext = baseListModel.hasNext() ? 
                        SimpleScalar.TRUE : SimpleScalar.FALSE;
                    SimpleNumber count = new SimpleNumber(indexCount);
                    modelRoot.put(hasNextName, hasNext);
                    modelRoot.put(counterName, count);
                    modelRoot.put(indexName, currElement);
                    block.process(modelRoot, out);
                    ++indexCount;
                }
            }
            else if (Template.CLASSIC_COMPATIBILITY && (baseModel instanceof TemplateScalarModel)) {
                modelRoot.put(hasNextName, SimpleScalar.FALSE);
                modelRoot.put(counterName, new SimpleNumber(0));
                modelRoot.put(indexName, baseModel);
                block.process(modelRoot, out);
            }
            else {
                String msg = "Error " + getLocation()
                             + "\nExpecting collection or list here"
                             + "\nbut " + listExpression + " is a "
                             + baseModel.getClass().getName() + ".";
                throw new TemplateException(msg);
            }
        }
        catch (BreakInstruction.Break br) {}
        finally {
            // Restore the old value of the loop index variable, if there was one.
            replace(modelRoot, indexName, oldIndexReferent);
            replace(modelRoot, hasNextName, prevHasNext);
            replace(modelRoot, counterName, prevCount);
        }
    }

    private void replace(TemplateModelRoot root, 
                         String keyName, 
                         TemplateModel keyValue) 
    {
        if (keyValue != null) {
            root.put(keyName, keyValue);
        } else {
            root.remove(keyName);
        }
    }

    public String toString() {
        if (isForEach) {
            StringBuffer buf = new StringBuffer("<foreach ");
            buf.append(indexName);
            buf.append(" in ");
            buf.append(listExpression);
            buf.append(">");
            buf.append(block);
            buf.append("</foreach>");
            return buf.toString();
        }
        else {
            StringBuffer buf = new StringBuffer("<list ");
            buf.append(listExpression);
            buf.append(" as ");
            buf.append(indexName);
            buf.append(">");
            buf.append(block);
            buf.append("</list>");
            return buf.toString();
        }
    }
}
