/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * 22 October 1999: This class added by Holger Arendt.
 */

package freemarker.template.compiler;

import freemarker.template.*;
import java.util.*;

/**
 * A unary operator that calls a TemplateMethodModel.  It associates with the
 * <tt>Identifier</tt> or <tt>Dot</tt> to its left.
 */
final class MethodCall extends Expression {

    private Expression target;
    private ListLiteral arguments;

    MethodCall(Expression target, List arguments) {
        this.target = target;
        this.arguments = new ListLiteral(arguments);
    }

    protected TemplateModel getAsTemplateModel(TemplateModelRoot modelRoot)
            throws TemplateException {

        TemplateModel targetModel = target.getAsTemplateModel(modelRoot);
        if (targetModel == null) {
            return null;
        }

        if (!(targetModel instanceof TemplateMethodModel)) {
            String msg = "Error " + getLocation()
                        + "\n" + target + " is not a TemplateMethodModel,"
                        + " it's a " + targetModel.getClass().getName() + ".";
            throw new TemplateException(msg);
        }

        TemplateMethodModel targetMethod = (TemplateMethodModel)targetModel;
        List argumentStrings = 
            targetMethod instanceof TemplateMethodModelEx
            ? arguments.getModelList(modelRoot)
            : arguments.getValueList(modelRoot);
        return targetMethod.exec(argumentStrings);
    }

    protected boolean isNumerical(TemplateModelRoot modelRoot) throws TemplateException {
        TemplateModel tm = getAsTemplateModel(modelRoot);
        return (tm instanceof TemplateNumberModel);
    }

    protected Number getNumericalValue(TemplateModelRoot modelRoot) throws TemplateException {
        TemplateModel tm = getAsTemplateModel(modelRoot);
        if (tm == null) {
            return null;
        }
        if (tm instanceof TemplateNumberModel) {
            return ((TemplateNumberModel) tm).getAsNumber();
        } 
        else {
            String msg = "Error " + getLocation()
                       + ".\nExpecting numerical value but method returned a "
                       + tm.getClass().getName() + ".";
            throw new NonNumericalException(msg);
        }
    }

    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append(target);
        buf.append("(");
        String list = arguments.toString();
        buf.append(list.substring(1, list.length() -1));
        buf.append(")");
        return buf.toString();
    }
}
