/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template.compiler;

import java.math.*;

/**
 * Utility methods for evaluating expressions.
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */

class NumberUtil {

    static int compareNumbers(Number first, Number second) {
        BigDecimal left = convertForArithmetic(first);
        BigDecimal right = convertForArithmetic(second);
        return left.compareTo(right);
    }

    static Number add(Number first, Number second) {
        BigDecimal left = convertForArithmetic(first);
        BigDecimal right = convertForArithmetic(second);
        return left.add(right);
    }

    static Number subtract(Number first, Number second) {
        BigDecimal left = convertForArithmetic(first);
        BigDecimal right = convertForArithmetic(second);
        return left.subtract(right);
    }

    static Number multiply(Number first, Number second) {
        BigDecimal left = convertForArithmetic(first);
        BigDecimal right = convertForArithmetic(second);
        BigDecimal result = left.multiply(right);
        if (result.scale() >100) {
            result = result.setScale(100, BigDecimal.ROUND_HALF_UP);
        }
        return result;
    }

    static Number divide(Number first, Number second) {
        BigDecimal left = convertForArithmetic(first);
        BigDecimal right = convertForArithmetic(second);
        return divide(left, right);
    }

    static Number modulus(Number first, Number second) {
        long left = first.longValue();
        long right = second.longValue();
        return new Long(left % right);
    }

    static private BigDecimal divide(BigDecimal left, BigDecimal right) {
        int scale1 = left.scale();
        int scale2 = right.scale();
        int scale = Math.max(scale1, scale2);
        scale = Math.max(MIN_SCALE, scale);
        return left.divide(right, scale, BigDecimal.ROUND_HALF_UP);
    }

    static private BigDecimal convertForArithmetic(Number num) {
        if (num instanceof BigDecimal) {
            return (BigDecimal) num;
        }
        return new BigDecimal(num.toString());
    }

    static private final int MIN_SCALE = 12;
}
