/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template.compiler;

import freemarker.template.*;
import java.io.Writer;
import java.io.IOException;
import java.math.BigDecimal;

/**
 * An instruction that outputs the value of a numerical expression.
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
final class NumericalOutput extends TemplateElement {

    private Expression expression;
    private int minSigDigits = -1;
    private int maxSigDigits = -1;

    NumericalOutput(Expression expression) {
        this.expression = expression;
    }

    NumericalOutput(Expression expression, int minSigDigits, int maxSigDigits) {
        this.expression = expression;
        this.minSigDigits = minSigDigits;
        this.maxSigDigits = maxSigDigits;
    }

    public void process(TemplateModelRoot modelRoot, Writer out)
        throws TemplateException, IOException
    {
        Number result = null;
        TemplateModel tm = expression.getAsTemplateModel(modelRoot);
        if (!(tm instanceof TemplateNumberModel)) {
            String msg = "Error " + getLocation();
            throw new NonNumericalException(msg);
        }
        Number num = ((TemplateNumberModel) tm).getAsNumber();
        String output = stripZeros(num.toString());
        if (minSigDigits >0 || maxSigDigits >=0) {
            output = adjustSigDigits(output, minSigDigits, maxSigDigits);
        }
        char decSep = LocaleUtil.getDecimalSeparator(modelRoot.getLocale());
        if (decSep != '.') {
            output = output.replace('.', decSep);
        }
        out.write(output);
    }

    static String stripZeros(String num) {
        int decind = num.indexOf('.');
        if (decind == -1) {
            return num;
        }
        int length = num.length();
        while (length > decind) {
            if (num.charAt(length-1) != '0') {
                break;
            }
            --length;
        }
        if (num.charAt(length-1) == '.') {
            --length;
        }
        num = num.substring(0, length);
        return num;
    }

    static String adjustSigDigits(String num, int minSigDigits, int maxSigDigits) {
        int decind = num.indexOf('.');
        String intPart, fraction;
        if (decind >=0) {
            intPart = num.substring(decind);
            fraction = num.substring(decind +1, num.length());
            if (fraction.length() > maxSigDigits) {
                BigDecimal bd = new BigDecimal(num);
                bd = bd.setScale(maxSigDigits, BigDecimal.ROUND_HALF_UP);
                num = bd.toString();
            }
            if (minSigDigits > fraction.length()) {
                StringBuffer buf = new StringBuffer(num);
                for (int i = fraction.length(); i< minSigDigits; i++) {
                    buf.append((char) '0');
                }
                num = buf.toString();
            }
        }
        else {
            if (minSigDigits >0) {
                StringBuffer buf = new StringBuffer(num);
                buf.append('.');
                for (int i = 0; i<minSigDigits; i++) {
                    buf.append((char) '0');
                }
                num = buf.toString();
            }
        }
        return num;
    }

    public String toString() {
        StringBuffer buf = new StringBuffer("#{");
        buf.append(expression);
        if (minSigDigits != -1 || maxSigDigits != -1) {
            buf.append(" ; ");
            if (minSigDigits != -1) {
                buf.append("m");
                buf.append(minSigDigits);
            }
            if (maxSigDigits != -1) {
                buf.append("M");
                buf.append(maxSigDigits);
            }
        }
        buf.append("}");
        return buf.toString();
    }
}
