/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template.compiler;

import freemarker.template.*;
import java.util.*;


/**
 * Some string utilities. 
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */

class StringUtil {

    static final String LITERALS  = "\"<>'&";

    static final String[] ENTITIES = new String[] {
        "&quot;",
        "&lt;",
        "&gt;",
        "&apos;",
        "&amp;"
    };


    /**
     * Converts the literal characters in a string
     * to HTML entities as need be.
     * @return converted string or simply echoes if no
     * conversion was necessary.
     */
    static String convertHTMLEntities(String s) {
        StringBuffer buf = new StringBuffer();
        boolean converted = false;
        for (int i = 0; i<s.length(); i++) {
            char c = s.charAt(i);
            int offset = LITERALS.indexOf(c);
            if (offset < 0) {
                buf.append(c);
            }
            else {
                buf.append(ENTITIES[offset]);
                converted = true;
            }
        }
        return converted ? buf.toString() : s;
    }


    /**
     * Convert the backslashes in a string literal.
     * (may need refinement)
     */
    static String convertEscapes(String s) {
       if (s.indexOf('\\') <0)
           return s;
       StringBuffer buf = new StringBuffer(); 
       StringTokenizer st = new StringTokenizer(s, "\\");
       boolean justAfterBackslash = false;
       while (st.hasMoreTokens()) {
           String tok = st.nextToken();
           if (justAfterBackslash) {
               char firstChar = tok.charAt(0);
               if (firstChar == '\\')
                  buf.append("\\");
               else if (firstChar == 'n')
                  buf.append("\n");
               else if (firstChar == 'r')
                  buf.append("\r");
               else if (firstChar == '\"')
                  buf.append("\"");
               else {
                  buf.append("\\"); 
                  buf.append(tok);
                  justAfterBackslash = false;
               }
               if (justAfterBackslash)
                  buf.append(tok.substring(1));
               justAfterBackslash = false;
           }
           else if (tok.equals("\\")) {
               justAfterBackslash = true;
           }
           else {
               buf.append(tok);
           }
       }
       return buf.toString();
    }

    static Locale deduceLocale(String input) {
       Locale locale = Locale.getDefault();
       if (input.charAt(0) == '"') input = input.substring(1, input.length() -1);
       StringTokenizer st = new StringTokenizer(input, ",_ ");
       String lang = "", country = "", variant="";
       if (st.hasMoreTokens()) {
          lang = st.nextToken();
       }
       if (st.hasMoreTokens()) {
          country = st.nextToken();
       }
       if (!st.hasMoreTokens()) {
          locale = new Locale(lang, country);
       } 
       else {
          locale = new Locale(lang, country, st.nextToken());
       }
       return locale;
    }

    static String capitalize(String s) {
        StringTokenizer st = new StringTokenizer(s, " \t\r\n", true);
        StringBuffer buf = new StringBuffer(s.length());
        while (st.hasMoreTokens()) {
            String tok = st.nextToken();
            buf.append(tok.substring(0, 1).toUpperCase());
            buf.append(tok.substring(1).toLowerCase());
        }
        return buf.toString();
    }
}
