/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template.compiler;

import freemarker.template.*;
import java.io.Writer;
import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;

/**
 * An instruction representing a switch-case structure.
 */
final class SwitchBlock extends TemplateElement {

    private LinkedList cases = new LinkedList();
    private Case defaultCase;
    private Expression testExpression;

    /**
     * @param testExpression the expression to be tested.
     */
    SwitchBlock(Expression testExpression) {
        this.testExpression = testExpression;
    }

    /**
     * @param cas a Case element.
     */
    void addCase(Case cas) {
        if (cas.isDefault) {
            defaultCase = cas;
        }
        else {
            cases.add(cas);
        }
    }

    public void process(TemplateModelRoot modelRoot, Writer out)
        throws TemplateException, IOException
    {
        boolean processedCase = false;
        Iterator iterator = cases.iterator();
        try {
            while (iterator.hasNext()) {
                Case cas = (Case)iterator.next();
                boolean processCase = false;

                // Fall through if a previous case tested true.
                if (processedCase) {
                    processCase = true;
                } else if (!cas.isDefault) {
                    // Otherwise, if this case isn't the default, test it.
                    ComparisonExpression equalsOp = new ComparisonExpression(testExpression, cas.expression, "==");
                    processCase = equalsOp.isTrue(modelRoot);
                }
                if (processCase) {
                    cas.process(modelRoot, out);
                    processedCase = true;
                }
            }

            // If we didn't process any cases, and we have a default,
            // process it.
            if (!processedCase && defaultCase != null) {
                defaultCase.process(modelRoot, out);
            }
        }
        catch (BreakInstruction.Break br) {}
    }

    public String toString() {
        StringBuffer buf = new StringBuffer("<switch ");
        buf.append(testExpression);
        buf.append(">");
        for (int i = 0; i<cases.size(); i++) {
            buf.append(cases.get(i));
        }
        if (defaultCase != null) {
            buf.append(defaultCase);
        }
        buf.append("</switch>");
        return buf.toString();
    }
}
