/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template.compiler;

import freemarker.template.*;
import java.io.*;

/**
 * An template element that contains a nested block
 * that is transformed according to an instance of T
 * TemplateTransformModel.
 */
final class TransformBlock extends TemplateElement {

    private Expression transformExpression;
    private TemplateElement content;

    /** Creates new TransformBlock, with a given transformation
     * @param variable the variable representing the transformation to use
     */
    TransformBlock( Expression transformExpression, TemplateElement content ) {
        this.transformExpression = transformExpression;
        this.content = content;
    }

    public void process(TemplateModelRoot modelRoot, Writer out) 
        throws TemplateException, IOException
    {
        TemplateTransformModel transform = null;
        try {
            transform = (TemplateTransformModel) transformExpression.getAsTemplateModel(modelRoot);
        }
        catch (ClassCastException cce) {
            String msg = "Error in instruction <transform " 
                 + transformExpression + ">\n"
                 + getLocation() + "\n"
                 + transformExpression
                 + " is not a TemplateTransformModel.";
            throw new TemplateException(msg, cce);
        }
        if (transform == null) {
            String msg = "Error in instruction <transform " 
                 + transformExpression + ">\n"
                 + getLocation() + "\n"
                 + transformExpression
                 + " is undefined.";
            throw new TemplateException(msg);
        }

        // Process the content into a string buffer.
        StringWriter sw = new StringWriter();
        try {
            // Get the content we have so far
            content.process(modelRoot, sw);
        }
        finally {
            Reader inData = new StringReader( sw.toString() );
            transform.transform(inData, out);
        }
    }

    public String toString() {
        StringBuffer buf = new StringBuffer("<transform ");
        buf.append(transformExpression);
        buf.append(">");
        buf.append(content);
        buf.append("</transform>");
        return buf.toString();
    }
}
