/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template.utility;

import org.apache.oro.text.perl.Perl5Util;
import org.apache.oro.text.regex.MatchResult;
import java.io.*;
import freemarker.template.TemplateTransformModel;
import freemarker.template.TemplateModelException;

/**
 * <p>A FreeMarker transformer that performs syntax colouring of FreeMarker
 * source files, and outputs the results in HTML. The HTML is HTML 3.2 compliant,
 * though later transitional DTDs should have no problems with this code.</p>
 *
 * <p>The transformation uses the Jakarta ORO regular expression library,
 * available from the <a href="http://jakarta.apache.org/oro/"
 * target="_top">Jakarta Apache</a> web site.</p>
 *
 * @author  Nicholas Cull
 * @version $Id: FreeMarkerToHtml.java,v 1.14 2002/04/21 13:49:31 revusky Exp $
 */
public class FreeMarkerToHtml implements TemplateTransformModel {
    private String textBuffer;
    private Perl5Util cMatcher;

    /** Creates new FreeMarkerToHtml */
    public FreeMarkerToHtml() {
        cMatcher = new Perl5Util();
    }

    /**
     * Perform all FreeMarker template transformations.
     */
    private void doFreeMarkerTransformation(String textBuffer, Writer out)
    throws IOException {
        textBuffer = cMatcher.substitute( "s/</&lt;/g", textBuffer );
        textBuffer = cMatcher.substitute( "s/>/&gt;/g", textBuffer );
//        textBuffer = cMatcher.substitute( "s/&/&amp;/g", textBuffer );
        out.write( "<pre>" );
        doBlockComment( textBuffer, out );
        out.write( "</pre>" );
    }

    /**
     * Format FreeMarker block comments.
     */
    private void doBlockComment(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/&lt;comment\\b(?:.*?)&lt;\\/comment\\s*&gt;/s",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            doFreeMarkerTags( aInput.substring( 0, cResult.beginOffset(0) ), out );
            out.write( "<font color=\"#008000\">" );
            out.write( cResult.toString() );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doFreeMarkerTags( aInput, out );
    }

    /**
     * Format FreeMarker start/standalone tags.
     */
    private void doFreeMarkerTags(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/(&lt;(?:list|if|else|elseif|switch|case|break|default|" +
                               "assign|include|function|call|compress|comment|noparse|foreach|" +
                               "transform))\\b(.*?)(&gt;)/s",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            doFreeMarkerEndTags( aInput.substring( 0, cResult.beginOffset(0) ), out );
            out.write( "<font color=\"#0000ff\">" );
            out.write( cResult.group( 1 ) );
            out.write( "</font>" );
            doDoubleQuotes( cResult.group( 2 ), out );
            out.write( "<font color=\"#0000ff\">" );
            out.write( cResult.group( 3 ) );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doFreeMarkerEndTags( aInput, out );
    }

    /**
     * Format FreeMarker end tags.
     */
    private void doFreeMarkerEndTags(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/(&lt;\\/(?:list|if|switch|" +
                               "function|compress|comment|noparse|foreach|" +
                               "transform))\\s*(&gt;)/s",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            doFreeMarkerVariableInstruction( aInput.substring( 0, cResult.beginOffset(0) ), out );
            out.write( "<font color=\"#0000ff\">" );
            out.write( cResult.group( 0 ) );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doFreeMarkerVariableInstruction( aInput, out );
    }

    /**
     * Format FreeMarker variable instructions.
     */
    private void doFreeMarkerVariableInstruction(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/([\\$\\#]\\{)(.*?)(\\})/s",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            out.write( aInput.substring( 0, cResult.beginOffset(0) ));
            out.write( "<font color=\"#0000ff\">" );
            out.write( cResult.group( 1 ) );
            out.write( "</font>" );
            doDoubleQuotes( cResult.group( 2 ), out );
            out.write( "<font color=\"#0000ff\">" );
            out.write( cResult.group( 3 ) );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        out.write( aInput );
    }

    /**
     * Format FreeMarker double-quoted strings.
     */
    private void doDoubleQuotes(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/\"(?:\\\\.|[^\"])*\"/m",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            doDigits( aInput.substring( 0, cResult.beginOffset(0) ), out);
            out.write( "<font color=\"#808080\">" );
            out.write( aMatch );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doDigits( aInput, out );
    }

    /**
     * Format FreeMarker number literals.
     */
    private void doDigits(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/\\b\\d+\\b/m",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            doOperator( aInput.substring( 0, cResult.beginOffset(0) ), out);
            out.write( "<font color=\"#ff0000\">" );
            out.write( cResult.toString() );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doOperator( aInput, out );
    }

    /**
     * Format FreeMarker operators.
     */
    private void doOperator(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/(?:[\\+\\[\\]\\{\\}\\=\\.\\(\\)\\;\\!]|&amp;)+/m",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            doSpecialWords( aInput.substring( 0, cResult.beginOffset(0) ), out);
            out.write( "<b>" );
            out.write( cResult.toString() );
            out.write( "</b>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doSpecialWords( aInput, out );
    }

    /**
     * Format FreeMarker special words, used in "foreach" and "list"
     * instructions.
     */
    private void doSpecialWords(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/\\b(?:as|in|type|gt|lt|gte|lte)\\b/m",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            doOtherWords( aInput.substring( 0, cResult.beginOffset(0) ), out );
            out.write( "<font color=\"#ff8000\">" );
            out.write( cResult.toString() );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doOtherWords( aInput, out );
    }

    /**
     * Format FreeMarker variables.
     */
    private void doOtherWords(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/\\w+/m",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            out.write( aInput.substring( 0, cResult.beginOffset(0) ));
            out.write( "<font color=\"#804040\">" );
            out.write( cResult.toString() );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        out.write( aInput );
    }

    /**
     * Performs a transformation/filter on FreeMarker output.
     *
     * @param source the input to be transformed
     * @param output the destination of the transformation
     */
    public void transform(Reader source, Writer output)
        throws TemplateModelException, IOException  
    {
        StringBuffer cText = new StringBuffer();
        char cBuffer[] = new char[ 1024 ];
        int nSize = source.read( cBuffer );
        while ( nSize >= 0 ) {
            cText.append( cBuffer, 0, nSize );
            nSize = source.read( cBuffer );
        }
        doFreeMarkerTransformation( cText.toString(), output );
    }

    /**
     * @return true if this object is empty.
     */
    public boolean isEmpty() {
        return false;
    }

}
