/*
 * FreeMarker: a tool that allows Java programs to use templates
 * to generate HTML or other text output that contains dynamic content.
 * Copyright (C) 1998,2002 Benjamin Geer
 * Email: beroul@users.sourceforge.net
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name "Freemarker" nor any of the names of the project 
 * contributors may be used to endorse or promote products 
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package freemarker.template.utility;

import org.apache.oro.text.perl.Perl5Util;
import org.apache.oro.text.regex.MatchResult;
import java.io.*;
import freemarker.template.TemplateTransformModel;
import freemarker.template.TemplateModelException;

/**
 * <p>A FreeMarker transformer that performs syntax colouring of Java source
 * files, and outputs the results in HTML. The HTML is HTML 3.2 compliant,
 * though later transitional DTDs should have no problems with this code.</p>
 *
 * <p>The transformation uses the Jakarta ORO regular expression library,
 * available from the <a href="http://jakarta.apache.org/oro/"
 * target="_top">Jakarta Apache</a> web site.</p>
 *
 * @author  Nicholas Cull
 * @version $Id: JavaToHtml.java,v 1.9 2002/04/23 10:35:37 szegedia Exp $
 */
public class JavaToHtml implements TemplateTransformModel {
    private Perl5Util cMatcher;

    /** Creates new JavaToHtml */
    public JavaToHtml() {
        cMatcher = new Perl5Util();
    }

    /**
     * Perform all Java language transformations.
     */
    private void doJavaTransformation( String textBuffer, Writer out ) throws IOException {
        textBuffer = cMatcher.substitute( "s/&/&amp;/g", textBuffer );
        textBuffer = cMatcher.substitute( "s/</&lt;/g", textBuffer );
        textBuffer = cMatcher.substitute( "s/>/&gt;/g", textBuffer );
        out.write( "<pre>" );
        doBlockComment( textBuffer, out );
        out.write( "</pre>" );
    }

    /**
     * Format Java block comments.
     */
    private void doBlockComment(String aInput, Writer out) 
    throws IOException 
    {
        while (cMatcher.match( "m/\\/\\*(?:.*?)\\*\\//s",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            doDoubleQuotes( aInput.substring( 0, cResult.beginOffset(0) ), out);
            out.write( "<font color=\"#008000\">" );
            out.write( cResult.toString() );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doDoubleQuotes( aInput, out );
    }

    /**
     * Format double quoted strings.
     */
    private void doDoubleQuotes(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/\"(?:\\\\.|[^\"])*\"/m",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            doSingleQuotes( aInput.substring( 0, cResult.beginOffset(0) ), out);
            out.write( "<font color=\"#808080\">" );
            out.write( aMatch );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doSingleQuotes( aInput, out );
    }

    /**
     * Format Java single quotes (ie. character data).
     */
    private void doSingleQuotes(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/'(?:\\\\.|[^\'])*'/m",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            doLineComment( aInput.substring( 0, cResult.beginOffset(0) ), out);
            out.write( "<font color=\"#808080\">" );
            out.write( cResult.toString() );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doLineComment( aInput, out );
    }

    /**
     * Format Java line comments (// syntax).
     */
    private void doLineComment(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/\\/\\/.*?$/m",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            doPrimitiveTypes( aInput.substring( 0, cResult.beginOffset(0) ), out);
            out.write( "<font color=\"#008000\">" );
            out.write( cResult.toString() );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doPrimitiveTypes( aInput, out );
    }

    /**
     * Format Java primitive types.
     */
    private void doPrimitiveTypes(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/\\b(?:boolean|byte|char|double|float|int|" +
                               "long|short|void)\\b/m",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            doKeywords( aInput.substring( 0, cResult.beginOffset(0) ), out);
            out.write( "<font color=\"#ff8000\">" );
            out.write( cResult.toString() );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doKeywords( aInput, out );
    }

    /**
     * Format Java keywords.
     */
    private void doKeywords(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/\\b(?:abstract|break|case|catch|" +
                               "class|continue|default|do|else|extends|" +
                               "false|final|finally|for|" +
                               "if|implements|import|instanceof|interface|length|" +
                               "native|new|null|package|private|protected|public|" +
                               "return|static|super|switch|synchronized|" +
                               "this|threadsafe|throw|throws|transient|true|try|" +
                               "volatile|while" +
                               ")\\b/m",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            doJavaPackages( aInput.substring( 0, cResult.beginOffset(0) ), out);
            out.write( "<font color=\"#0000ff\">" );
            out.write( cResult.toString() );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doJavaPackages( aInput, out );
    }

    /**
     * Format three of the main Java classes: Java.lang, Java.io, and Java.util.
     */
    private void doJavaPackages(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/\\b(?:AbstractCollection|AbstractList|AbstractMap|" +
                               "AbstractSequentialList|AbstractSet|ArrayList|Arrays|" +
                               "BitSet|Boolean|BufferedInputStream|" +
                               "BufferedOutputStream|BufferedReader|BufferedWriter|Byte|" +
                               "ByteArrayInputStream|ByteArrayOutputStream|Calendar|Character|" +
                               "Class|CharArrayReader|CharArrayWriter|ClassLoader|Cloneable|" +
                               "Collection|Collections|Comparable|Comparator|" +
                               "Compiler|DataInput|DataInputStream|DataOutput|DataOutputStream|" +
                               "Date|Dictionary|Double|Enumeration|EventListener|EventObject|" +
                               "Externalizable|File|FileDescriptor|FileFilter|FilenameFilter|" +
                               "FileInputStream|FileOutputStream|FilePermission|FileReader|" +
                               "FileWriter|FilterInputStream|FilterOutputStream|FilterReader|" +
                               "FilterWriter|Float|GregorianCalendar|HashMap|HashSet|Hashtable|" +
                               "InheritableThreadLocal|InputStream|InputStreamReader|Integer|" +
                               "Iterator|LineNumberInputStream|LineNumberReader|LinkedList|" +
                               "List|ListIterator|ListResourceBundle|Locale|Long|Map|Math|" +
                               "Number|Object|ObjectInput|ObjectInputStream|ObjectInputValidation|" +
                               "ObjectOutput|ObjectOutputStream|ObjectStreamClass|" +
                               "ObjectStreamConstants|ObjectStreamField|Observable|Observer|" +
                               "OutputStream|OutputStreamWriter|Package|PipedInputStream|" +
                               "PipedOutputStream|PipedReader|PipedWriter|PrintStream|" +
                               "Writer|Process|Properties|PropertyPermission|" +
                               "PropertyResourceBundle|PushbackInputStream|PushbackReader|" +
                               "Random|RandomAccessFile|Reader|ResourceBundle|Runnable|" +
                               "Runtime|RuntimePermission|SecurityManager|SequenceInputStream|" +
                               "Serializable|SerializablePermission|Set|Short|SimpleTimeZone|" +
                               "SortedMap|SortedSet|Stack|StreamTokenizer|StrictMath|String|" +
                               "StringBuffer|StringBufferInputStream|StringReader|" +
                               "StringTokenizer|StringWriter|System|Thread|ThreadGroup|" +
                               "ThreadLocal|Timer|TimerTask|TimeZone|TreeMap|TreeSet|" +
                               "Vector|Void|WeakHashMap|Writer" +
                               ")\\b/m",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            doDigits( aInput.substring( 0, cResult.beginOffset(0) ), out);
            out.write( "<font color=\"#804040\">" );
            out.write( cResult.toString() );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doDigits( aInput, out );
    }

    /**
     * Format digits in Java source code.
     */
    private void doDigits(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/\\b\\d+\\b/m",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            doOperator( aInput.substring( 0, cResult.beginOffset(0) ), out);
            out.write( "<font color=\"#ff0000\">" );
            out.write( cResult.toString() );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doOperator( aInput, out );
    }

    /**
     * Format Java operators.
     */
    private void doOperator(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/(?:\\w+\\s*\\(|[\\+\\*\\^\\$\\-\\{\\}\\[\\]" +
                               "\\=\\.\\(\\)\\,\\:\\/\\;]|&(lt|gt|amp);" +
                               ")+/m",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            doExceptions( aInput.substring( 0, cResult.beginOffset(0) ), out);
            out.write( "<b>" );
            out.write( cResult.toString() );
            out.write( "</b>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        doExceptions( aInput, out );
    }

    /**
     * Format Java exception classes.
     */
    private void doExceptions(String aInput, Writer out)  throws IOException {
        while (cMatcher.match( "m/\\w*(Error|Exception|Throwable)\\b/m",  aInput)) {
            MatchResult cResult = cMatcher.getMatch();
            String aMatch = cResult.toString();
            out.write( aInput.substring( 0, cResult.beginOffset(0) ));
            out.write( "<font color=\"#ff0000\">" );
            out.write( cResult.toString() );
            out.write( "</font>" );
            aInput = aInput.substring( cResult.endOffset(0) );
        }
        out.write( aInput );
    }

    /**
     * Performs a transformation/filter on FreeMarker output.
     *
     * @param source the input to be transformed
     * @param output the destination of the transformation
     */
    public void transform(Reader source, Writer output)
    throws TemplateModelException, IOException {
        StringBuffer cText = new StringBuffer();
        char cBuffer[] = new char[ 1024 ];
        int    nSize = source.read( cBuffer );
        while ( nSize >= 0 ) {
            cText.append( cBuffer, 0, nSize );
            nSize = source.read( cBuffer );
        }
        doJavaTransformation( cText.toString(), output );
    }

    /**
     * @return true if this object is empty.
     */
    public boolean isEmpty() {
        return false;
    }

}
